# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module docstring."""
from typing import Any, Dict, Optional, Union

import ipywidgets as widgets
from IPython.display import display

from .._version import VERSION
from .ce_azure_sentinel import CEAzureSentinel
from .ce_data_providers import CEDataProviders
from .ce_keyvault import CEKeyVault
from .ce_azure import CEAzure
from .ce_other_providers import CEOtherProviders
from .ce_ti_providers import CETIProviders
from .ce_user_defaults import CEAutoLoadComps, CEAutoLoadQProvs
from .comp_edit import CompEditDisplayMixin, CompEditTabs, CETabControlDef
from .mp_config_file import MpConfigFile
from .mp_config_control import MpConfigControls, get_mpconfig_definitions

__version__ = VERSION
__author__ = "Ian Hellen"


class MpConfigEdit(CompEditDisplayMixin):
    """Msticpy Configuration helper class."""

    _TAB_DEFINITIONS = {
        "MicrosoftSentinel": CEAzureSentinel,
        "TI Providers": CETIProviders,
        "Data Providers": CEDataProviders,
        "GeoIP Providers": CEOtherProviders,
        "Key Vault": CEKeyVault,
        "Azure": CEAzure,
        "Autoload QueryProvs": CEAutoLoadQProvs,
        "Autoload Components": CEAutoLoadComps,
    }

    def __init__(
        self,
        settings: Optional[Union[Dict[str, Any], MpConfigFile, str]] = None,
        conf_filepath: str = None,
    ):
        """
        Initialize instance of MpConfigEdit.

        Parameters
        ----------
        settings : Optional[Union[Dict[str, Any], MpConfigFile, str]], optional
            settings can be one of the following:

            - A dict of msticpyconfig settings
            - An instance of MpConfigFile with loaded settings
            - A file path to a msticpyconfig.yaml

            If None, the settings will be read from the
            default (via MSTICPYCONFIG variable)

        conf_filepath : str
            If settings are passed as MPConfigFile instance or a dict,
            this parameter will override the file path used to save the
            settings.
            If settings is a file path, this parameter is ignored.

        """
        self._lbl_loading = widgets.Label(value="Loading. Please wait.")
        display(self._lbl_loading)
        if isinstance(settings, MpConfigFile):
            self.mp_conf_file = MpConfigFile(settings=settings.settings)
            if not self.mp_conf_file.current_file and conf_filepath:
                self.mp_conf_file.current_file = conf_filepath
        elif isinstance(settings, dict):
            self.mp_conf_file = MpConfigFile(settings=settings)
            if not self.mp_conf_file.current_file and conf_filepath:
                self.mp_conf_file.current_file = conf_filepath
        elif isinstance(settings, str):
            self.mp_conf_file = MpConfigFile()
            self.mp_conf_file.load_from_file(file=settings)
        else:
            self.mp_conf_file = MpConfigFile()
            self.mp_conf_file.load_default()
        self.tool_buttons: Dict[str, widgets.Widget] = {}
        self._inc_loading_label()

        # Get the settings definitions and Config controls object
        mp_def_dict = get_mpconfig_definitions()
        self.mp_controls = MpConfigControls(mp_def_dict, self.mp_conf_file.settings)
        self._inc_loading_label()

        # Set up the tabs
        self.tab_ctrl = CompEditTabs(self._get_tab_definitions())
        self._inc_loading_label()

        self.txt_current_file = widgets.Text(
            description="Conf File",
            value=self.current_config_file,
            layout=widgets.Layout(width="75%"),
        )
        self.btn_save = widgets.Button(
            description="Save Settings",
            tooltip="Save current settings to your config file.",
        )
        self.btn_save.on_click(self._save_file)
        self.btn_validate = widgets.Button(
            description="Validate Settings",
            tooltip="Run basic sanity checks on current settings.",
        )
        self.btn_validate.on_click(self._validate_config)
        self.cb_backup = widgets.Checkbox(description="Create backup", value=False)
        vbox = widgets.VBox(
            [
                self.txt_current_file,
                widgets.HBox([self.btn_save, self.cb_backup, self.btn_validate]),
                self.mp_conf_file.viewer,
            ]
        )
        self.layout = widgets.VBox([self.tab_ctrl.layout, vbox])
        self._lbl_loading.layout.visibility = "hidden"

    def _inc_loading_label(self):
        self._lbl_loading.value = f"{self._lbl_loading.value}."

    @property
    def tab_names(self):
        """Return a list of current tabs."""
        return self.tab_ctrl.tab_names

    @property
    def controls(self):
        """Return a list of current tab names and controls."""
        return self.tab_ctrl.tab_controls

    def set_tab(self, tab_name: Optional[str], index: int = 0):
        """Programatically set the tab by name or index."""
        self.tab_ctrl.set_tab(tab_name, index)

    def _save_file(self, btn):
        del btn
        if self.txt_current_file.value:
            self.mp_conf_file.save_to_file(
                self.txt_current_file.value, backup=self.cb_backup.value
            )

    def _validate_config(self, btn):
        del btn
        self.mp_conf_file.validate_settings()
        mpc_validation = self.mp_controls.validate_all_settings()
        log_val_txt = self.mp_conf_file.txt_viewer.value
        self.mp_conf_file.txt_viewer.value = "\n".join(
            [
                "Logical validation results",
                "--------------------------",
                log_val_txt,
                "",
                "Type validation results",
                "-----------------------",
                *(item.status for item in mpc_validation),
            ]
        )

    def _create_data_tabs(self):
        """Create all tab contents."""
        self.tab_ctrl.tab.children = []
        for name, (ctrl_cls, args) in self._get_tab_definitions().items():
            ctrl = ctrl_cls(*args)
            # add to tabs
            self.tab_ctrl.add_tab(name, control=ctrl)
            # Set these controls as named attributes on the object
            setattr(self, name.replace(" ", "_"), ctrl)

    def _get_tab_definitions(self) -> Dict[str, CETabControlDef]:
        """Return tab definitions and arguments."""
        return {
            name: (cls, [self.mp_controls])
            for name, cls in self._TAB_DEFINITIONS.items()
        }

    @property
    def current_config_file(self):
        """Return the currently loaded configuration file path."""
        return self.mp_conf_file.current_file



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MailCluster Entity class."""
from typing import Any, Dict, List, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name, too-many-instance-attributes


@export
class MailCluster(Entity):
    """
    MailCluster Entity class.

    Attributes
    ----------
    NetworkMessageIds : List[str]
        NetworkMessageIds of the Mail cluster
    CountByDeliveryStatus : Dict[str, int] = {}
        CountByDeliveryStatus of the Mail cluster
    CountByThreatType : Dict[str, int] = {}
        CountByThreatType of the Mail cluster
    CountByProtectionStatus : Dict[str, int] = {}
        CountByProtectionStatus of the Mail cluster
    Threats : List[str]
        Threats of the Mail cluster
    Query : str
        Query of the Mail cluster
    QueryTime : datetime
        QueryTime of the Mail cluster
    MailCount : int
        MailCount of the Mail cluster
    IsVolumeAnomaly : bool
        IsVolumeAnomaly of the Mail cluster
    Source : str
        Source of the Mail cluster
    ClusterSourceIdentifier : str
        ClusterSourceIdentifier of the Mail cluster
    ClusterSourceType : str
        ClusterSourceType of the Mail cluster
    ClusterQueryStartTime : datetime
        ClusterQueryStartTime of the Mail cluster
    ClusterQueryEndTime : datetime
        ClusterQueryEndTime of the Mail cluster
    ClusterGroup : str
        ClusterGroup of the Mail cluster

    """

    ID_PROPERTIES = ["Query", "Source"]

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.NetworkMessageIds: List[str] = []
        self.CountByDeliveryStatus: Dict[str, int] = {}
        self.CountByThreatType: Dict[str, int] = {}
        self.CountByProtectionStatus: Dict[str, int] = {}
        self.Threats: List[str] = []
        self.Query: Optional[str] = None
        self.QueryTime: Any = None
        self.MailCount: int = 0
        self.IsVolumeAnomaly: bool = False
        self.Source: Optional[str] = None
        self.ClusterSourceIdentifier: Optional[str] = None
        self.ClusterSourceType: Optional[str] = None
        self.ClusterQueryStartTime: Any = None
        self.ClusterQueryEndTime: Any = None
        self.ClusterGroup: Optional[str] = None

        super().__init__(src_entity=src_entity, **kwargs)
        if src_event is not None:
            self._create_from_event(src_event)

    def _create_from_event(self, src_event):
        self.NetworkMessageIds = src_event["NetworkMessageIds"]
        self.CountByThreatType = src_event["CountByThreatType"]
        self.CountByProtectionStatus = src_event["CountByProtectionStatus"]
        self.Query = src_event["Query"]
        self.QueryTime = src_event["QueryTime"]
        self.MailCount = src_event["MailCount"]
        self.Source = src_event["Source"]

    @property
    def description_str(self):
        """Return Entity Description."""
        return self.Query or self.NetworkMessageIds or self.__class__.__name__

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        hash_val = hash(str(self.NetworkMessageIds))
        return f"{self.__class__.__name__} - {hash_val}"

    _entity_schema = {
        "NetworkMessageIds": None,
        "CountByDeliveryStatus": None,
        "CountByThreatType": None,
        "CountByProtectionStatus": None,
        "Threats": None,
        "Query": None,
        "QueryTime": None,
        "MailCount": None,
        "IsVolumeAnomaly": None,
        "Source": None,
        "ClusterSourceIdentifier": None,
        "ClusterSourceType": None,
        "ClusterQueryStartTime": None,
        "ClusterQueryEndTime": None,
        "ClusterGroup": None,
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Dns Entity class."""
from typing import Any, List, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity
from .ip_address import IpAddress

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name


@export
class Dns(Entity):
    """
    DNS Resolve Entity class.

    Attributes
    ----------
    DomainName : str
        DnsResolve DomainName
    IpAdresses : List[str]
        DnsResolve IpAdresses
    DnsServerIp : IPAddress
        DnsResolve DnsServerIp
    HostIpAddress : IPAddress
        DnsResolve HostIpAddress

    """

    ID_PROPERTIES = ["DomainName"]

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.DomainName: Optional[str] = None
        self.IpAdresses: List[IpAddress] = []
        self.DnsServerIp: Optional[IpAddress] = None
        self.HostIpAddress: Optional[IpAddress] = None
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.DomainName}: IPs: {repr(self.IpAdresses)}"

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.DomainName or self.__class__.__name__

    _entity_schema = {
        # DomainName (type System.String)
        "DomainName": None,
        # IpAdresses (type System.Collections.Generic.List`1
        # [Microsoft.Azure.Security.Detection.AlertContracts.V3.Entities.IP])
        "IpAdresses": (List, "IpAddress"),
        # DnsServerIp (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.IP)
        "DnsServerIp": "IpAddress",
        # HostIpAddress (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.IP)
        "HostIpAddress": "IpAddress",
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
HTTP TI Provider base.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
import abc
import traceback
from functools import lru_cache
from http import client
from json import JSONDecodeError
from typing import Any, Dict, List, Tuple

import attr
import requests
from attr import Factory

from ..._version import VERSION
from ...common.exceptions import MsticpyConfigException
from ...common.utility import export, _MSTICPY_USER_AGENT
from .ti_provider_base import LookupResult, TILookupStatus, TIProvider, TISeverity

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True)
class IoCLookupParams:
    """IoC HTTP Lookup Params definition."""

    path: str = ""
    verb: str = "GET"
    full_url: bool = False
    headers: Dict[str, str] = Factory(dict)
    params: Dict[str, str] = Factory(dict)
    data: Dict[str, str] = Factory(dict)
    auth_type: str = ""
    auth_str: List[str] = Factory(list)
    sub_type: str = ""


@export
class HttpProvider(TIProvider):
    """HTTP TI provider base class."""

    _BASE_URL = ""

    _IOC_QUERIES: Dict[str, IoCLookupParams] = {}

    _REQUIRED_PARAMS: List[str] = []

    def __init__(self, **kwargs):
        """Initialize a new instance of the class."""
        super().__init__(**kwargs)

        self._requests_session = requests.Session()
        self._request_params = {}
        if "ApiID" in kwargs:
            self._request_params["API_ID"] = kwargs.pop("ApiID")
        if "AuthKey" in kwargs:
            self._request_params["API_KEY"] = kwargs.pop("AuthKey")

        missing_params = [
            param
            for param in self._REQUIRED_PARAMS
            if param not in self._request_params
        ]
        if missing_params:
            param_list = ", ".join(f"'{param}'" for param in missing_params)
            raise MsticpyConfigException(
                f"Parameter values missing for TI Provider '{self.__class__.__name__}'",
                f"Missing parameters are: {param_list}",
            )

    # pylint: disable=too-many-branches, duplicate-code
    @lru_cache(maxsize=256)
    def lookup_ioc(  # type: ignore
        self, ioc: str, ioc_type: str = None, query_type: str = None, **kwargs
    ) -> LookupResult:
        """
        Lookup a single IoC observable.

        Parameters
        ----------
        ioc : str
            IoC observable
        ioc_type : str, optional
            IocType, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.

        Returns
        -------
        LookupResult
            The lookup result:
            result - Positive/Negative,
            details - Lookup Details (or status if failure),
            raw_result - Raw Response
            reference - URL of IoC

        Raises
        ------
        NotImplementedError
            If attempting to use an HTTP method or authentication
            protocol that is not supported.

        Notes
        -----
        Note: this method uses memoization (lru_cache) to cache results
        for a particular observable to try avoid repeated network calls for
        the same item.

        """
        result = self._check_ioc_type(
            ioc=ioc, ioc_type=ioc_type, query_subtype=query_type
        )

        result.provider = kwargs.get("provider_name", self.__class__.__name__)
        if result.status:
            return result

        req_params: Dict[str, Any] = {}
        try:
            verb, req_params = self._substitute_parms(
                result.safe_ioc, result.ioc_type, query_type
            )
            if verb == "GET":
                response = self._requests_session.get(**req_params)
            else:
                raise NotImplementedError(f"Unsupported verb {verb}")
            result.status = response.status_code
            result.reference = req_params["url"]
            if result.status == 200:
                try:
                    result.raw_result = response.json()
                    result.result, severity, result.details = self.parse_results(result)
                except JSONDecodeError:
                    result.raw_result = f"""There was a problem parsing results from this lookup:
                                        {response.text}"""
                    result.result = False
                    severity = TISeverity.information
                    result.details = {}
                result.set_severity(severity)
                result.status = TILookupStatus.ok.value
            else:
                result.raw_result = str(response)
                result.result = False
                result.details = self._response_message(result.status)
            return result
        except (  # pylint: disable=duplicate-code
            LookupError,
            JSONDecodeError,
            NotImplementedError,
            ConnectionError,
        ) as err:
            self._err_to_results(result, err)
            if not isinstance(err, LookupError):
                url = req_params.get("url", None) if req_params else None
                result.reference = url
            return result

    # pylint: enable=duplicate-code
    # pylint: disable=too-many-branches
    def _substitute_parms(
        self, ioc: str, ioc_type: str, query_type: str = None
    ) -> Tuple[str, Dict[str, Any]]:
        """
        Create requests parameters collection.

        Parameters
        ----------
        ioc : str
            IoC observable
        ioc_type : str, optional
            IocType, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.

        Returns
        -------
        Tuple[str, Dict[str, Any]]
            HTTP method, dictionary of parameter keys/values

        """
        req_params = {"observable": ioc}
        req_params.update(self._request_params)
        ioc_key = ioc_type + "-" + query_type if query_type else ioc_type
        src = self._IOC_QUERIES.get(ioc_key, None)
        if not src:
            raise LookupError(f"Provider does not support IoC type {ioc_key}.")

        # create a parameter dictionary to pass to requests
        # substitute any parameter value from our req_params dict
        req_dict: Dict[str, Any] = {
            "headers": {},
            "url": self._BASE_URL + src.path.format(**req_params)
            if not src.full_url
            else src.path.format(observable=ioc),
        }

        if src.headers:
            headers: Dict[str, Any] = {
                key: val.format(**req_params) for key, val in src.headers.items()
            }
            req_dict["headers"] = headers
        if "User-Agent" not in req_dict["headers"]:
            req_dict["headers"]["User-Agent"] = _MSTICPY_USER_AGENT
        if src.params:
            q_params: Dict[str, Any] = {
                key: val.format(**req_params) for key, val in src.params.items()
            }
            req_dict["params"] = q_params
        if src.data:
            q_data: Dict[str, Any] = {
                key: val.format(**req_params) for key, val in src.data.items()
            }
            req_dict["data"] = q_data
        if src.auth_type and src.auth_str:
            auth_strs: Tuple = tuple(p.format(**req_params) for p in src.auth_str)
            if src.auth_type == "HTTPBasic":
                req_dict["auth"] = auth_strs
            else:
                raise NotImplementedError(f"Unknown auth type {src.auth_type}")
        return src.verb, req_dict

    @abc.abstractmethod
    def parse_results(self, response: LookupResult) -> Tuple[bool, TISeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : LookupResult
            The returned data response

        Returns
        -------
        Tuple[bool, TISeverity, Any]
            bool = positive or negative hit
            TISeverity = enumeration of severity
            Object with match details

        """

    @staticmethod
    def _failed_response(response: LookupResult) -> bool:
        """
        Return True if negative response.

        Parameters
        ----------
        response : LookupResult
            The returned data response

        Returns
        -------
        bool
            True if the response indicated failure.

        """
        return (
            response.status != 200
            or not response.raw_result
            or not isinstance(response.raw_result, dict)
        )

    @staticmethod
    def _err_to_results(result: LookupResult, err: Exception):
        result.details = err.args
        result.raw_result = (
            type(err).__name__ + "\n" + str(err) + "\n" + traceback.format_exc()
        )

    @staticmethod
    def _response_message(status_code):
        if status_code == 404:
            return "Not found."
        if status_code == 401:
            return "Authorization failed. Check account and key details."
        if status_code == 403:
            return "Request forbidden. Allowed query rate may have been exceeded."
        return client.responses.get(status_code, "Unknown HTTP status code.")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Helper module for laplace smoothing counts."""

from typing import Tuple, List, DefaultDict
import copy


def laplace_smooth_cmd_counts(
    seq1_counts: DefaultDict[str, int],
    seq2_counts: DefaultDict[str, DefaultDict[str, int]],
    start_token: str,
    end_token: str,
    unk_token: str,
) -> Tuple[DefaultDict[str, int], DefaultDict[str, DefaultDict[str, int]]]:
    """
    Apply laplace smoothing to the input counts for the cmds.

    In particular, add 1 to each of the counts, including the unk_token. By including the
    unk_token, we can handle unseen commands.

    Parameters
    ----------
    seq1_counts: DefaultDict[str, int]
        individual command counts
    seq2_counts: DefaultDict[str, DefaultDict[str, int]]
        sequence command (length 2) counts
    start_token: str
        dummy command to signify the start of a session (e.g. "##START##")
    end_token: str
        dummy command to signify the end of a session (e.g. "##END##")
    unk_token: str
        dummy command to signify an unseen command (e.g. "##UNK##")

    Returns
    -------
    tuple of laplace smoothed counts:
        individual command counts,
        sequence command (length 2) counts

    """
    seq1_counts_ls = copy.deepcopy(seq1_counts)
    seq2_counts_ls = copy.deepcopy(seq2_counts)

    cmds: List[str] = list(seq1_counts_ls.keys()) + [unk_token]
    for cmd1 in cmds:
        for cmd2 in cmds:
            if cmd1 != end_token and cmd2 != start_token:
                seq1_counts_ls[cmd1] += 1
                seq2_counts_ls[cmd1][cmd2] += 1
                seq1_counts_ls[cmd2] += 1

    return seq1_counts_ls, seq2_counts_ls


def laplace_smooth_param_counts(
    cmds: List[str],
    param_counts: DefaultDict[str, int],
    cmd_param_counts: DefaultDict[str, DefaultDict[str, int]],
    unk_token: str,
) -> Tuple[DefaultDict[str, int], DefaultDict[str, DefaultDict[str, int]]]:
    """
    Apply laplace smoothing to the input counts for the params.

    In particular, add 1 to each of the counts, including the unk_token. By including the
    unk_token, we can handle unseen params.

    Parameters
    ----------
    cmds: List[str]
        list of all the possible commands (including the unk_token)
    param_counts: DefaultDict[str, int]
        individual param counts
    cmd_param_counts: DefaultDict[str, DefaultDict[str, int]]
        param conditional on command counts
    unk_token: str
        dummy command to signify an unseen command (e.g. "##UNK##")

    Returns
    -------
    Tuple:
        individual param probabilities,
        param conditional on command probabilities

    """
    param_counts_ls = copy.deepcopy(param_counts)
    cmd_param_counts_ls = copy.deepcopy(cmd_param_counts)

    params: List[str] = list(param_counts.keys()) + [unk_token]
    for cmd in cmds:
        for param in params:
            if param in cmd_param_counts_ls[cmd] or param == unk_token:
                param_counts_ls[param] += 1
                cmd_param_counts_ls[cmd][param] += 1

    return param_counts_ls, cmd_param_counts_ls


def laplace_smooth_value_counts(
    params: List[str],
    value_counts: DefaultDict[str, int],
    param_value_counts: DefaultDict[str, DefaultDict[str, int]],
    unk_token: str,
) -> Tuple[DefaultDict[str, int], DefaultDict[str, DefaultDict[str, int]]]:
    """
    Apply laplace smoothing to the input counts for the values.

    In particular, add 1 to each of the counts, including the unk_token. By including the
    unk_token, we can handle unseen values.

    Parameters
    ----------
    params: List[str]
        list of all possible params, including the unk_token
    value_counts: DefaultDict[str, int]
        individual value counts
    param_value_counts: DefaultDict[str, DefaultDict[str, int]]
        value conditional on param counts
    unk_token: str
        dummy command to signify an unseen command (e.g. "##UNK##")

    Returns
    -------
    Tuple:
        individual value probabilities,
        value conditional on param probabilities

    """
    value_counts_ls = copy.deepcopy(value_counts)
    param_value_counts_ls = copy.deepcopy(param_value_counts)

    values: List[str] = list(value_counts_ls.keys()) + [unk_token]
    for param in params:
        for value in values:
            if value in param_value_counts_ls[param] or value == unk_token:
                value_counts_ls[value] += 1
                param_value_counts_ls[param][value] += 1

    return value_counts_ls, param_value_counts_ls



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MSTIC Anomalous Sequence Modelling Tools."""

from ..._version import VERSION

__version__ = VERSION



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""GeoLocation Entity class."""
from typing import Any, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import ContextObject, Entity

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name


@export
class GeoLocation(Entity, ContextObject):
    """
    GeoLocation class.

    Attributes
    ----------
    CountryCode : str
        GeoLocation CountryCode
    CountryName : str
        GeoLocation CountryName
    State : str
        GeoLocation State
    City : str
        GeoLocation City
    Longitude : float
        GeoLocation Longitude
    Latitude : float
        GeoLocation Latitude
    Asn : str
        GeoLocation Asn

    """

    ID_PROPERTIES = ["Longitude", "Latitude", "City", "State", "CountryCode"]

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.CountryCode: Optional[str] = None
        self.CountryName: Optional[str] = None
        self.State: Optional[str] = None
        self.City: Optional[str] = None
        self.Longitude: Optional[str] = None
        self.Latitude: Optional[str] = None
        self.Asn: Optional[str] = None
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.CountryCode}; {self.State}; {self.City}"

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.CountryCode or self.__class__.__name__

    _entity_schema = {
        # str
        "CountryCode": None,
        # str
        "CountryName": None,
        # str
        "State": None,
        # str
        "City": None,
        # double?
        "Longitude": None,
        # double?
        "Latitude": None,
        # int
        "Asn": None,
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MSTIC Security Tools."""

# flake8: noqa: F403
# pylint: disable=W0401
from .iocextract import IoCExtract
from .geoip import GeoLiteLookup, IPStackLookup, geo_distance
from .tilookup import TILookup
from .vtlookup import VTLookup
from . import base64unpack as base64
from . import process_tree_utils as ptree
from .._version import VERSION

try:
    from IPython import get_ipython
    from . import sectools_magics
except ImportError as err:
    pass

__version__ = VERSION



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Malware Entity class."""
from typing import Any, List, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity
from .file import File
from .process import Process

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name


@export
class Malware(Entity):
    """
    Malware Entity class.

    Attributes
    ----------
    Name : str
        Malware Name
    Category : str
        Malware Category
    File : File
        Malware File
    Files : List[File]
        Malware Files
    Processes : List[Process]
        Malware Processes

    """

    ID_PROPERTIES = ["Name"]

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.Name: str = ""
        self.Category: str = ""
        self.File: Optional[File] = None
        self.Files: List[File] = []
        self.Processes: List[Process] = []
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.Name}: {self.Category}"

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.Name or self.__class__.__name__

    _entity_schema = {
        # Name (type System.String)
        "Name": None,
        # Category (type System.String)
        "Category": None,
        # File (type Microsoft.Azure.Security.Detection.AlertContracts.V3.Entities.File)
        "File": "File",
        "Files": (list, "File"),
        "Processes": (list, "Process"),
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Pivot query functions class."""
import itertools
import warnings
from collections import defaultdict, namedtuple, abc
from functools import wraps
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Type

import pandas as pd

from .pivot_register import join_result, get_join_params
from ..common.timespan import TimeSpan
from .._version import VERSION
from ..data.data_providers import QueryProvider
from ..data.query_container import QueryContainer
from ..data.query_source import QuerySource
from . import entities

__version__ = VERSION
__author__ = "Ian Hellen"


ParamAttrs = namedtuple("ParamAttrs", "type, query, family, required")
QueryParams = namedtuple(
    "QueryParams", "all, required, full_required, param_attrs, table"
)
PivQuerySettings = namedtuple(
    "PivQuerySettings", "short_name, direct_func_entities, assigned_entities"
)

_DEF_IGNORE_PARAM = {"start", "end"}

_TABLE_SHORTNAMES = {
    "SecurityEvent": "wevt",
    "Syslog": "lxsys",
    "SecurityAlert": "",
    "SigninLogs": "aad",
    "AzureActivity": "az",
    "AzureNetworkAnalytics_CL": "aznet",
    "OfficeActivity": "o365",
    "ThreatIntelligenceIndicator": "azti",
    "Heartbeat": "hb",
    "AuditLog_CL": "lxaud",
    "HuntingBookmark": "azsent",
    "StorageFileLogs": "az",
    "DnsEvents": "dns",
}


class PivotQueryFunctions:
    """Class to retrieve the queries and params from a provider."""

    current = None

    def __init__(self, query_provider: QueryProvider, ignore_reqd: List[str] = None):
        """
        Instantiate PivotQueryFunctions class.

        Parameters
        ----------
        query_provider : [type]
            The query provider to load
        ignore_reqd : List[str], optional
            List of parameters to ignore when building the required
            parameters list (e.g. ['start', 'end']), by default None

        """
        self.__class__.current = self
        self._provider = query_provider
        self.param_usage: Dict[str, List[ParamAttrs]] = defaultdict(list)
        self.query_params: Dict[str, QueryParams] = {}

        # specify any parameters to exclude from our list
        ignore_params = set(ignore_reqd) if ignore_reqd else _DEF_IGNORE_PARAM

        # get the query dict for each data family
        for family, fam_dict in self._provider.query_store.data_families.items():
            # for each query
            for src_name, q_source in fam_dict.items():
                # get the set of required params
                reqd_params = set(q_source.required_params.keys()) - ignore_params
                # add them to the param_usage attrib
                for param, p_attrs in q_source.params.items():
                    self.param_usage[param].append(
                        ParamAttrs(
                            p_attrs["type"],
                            src_name,
                            family,
                            bool(param in reqd_params),
                        )
                    )
                # add an entry to the query dictionary containing full
                # details of the function/query parameters
                self.query_params[f"{family}.{src_name}"] = QueryParams(
                    all=list(q_source.params),
                    required=list((set(q_source.required_params) - ignore_params)),
                    full_required=list(q_source.required_params),
                    param_attrs={
                        param: ParamAttrs(
                            p_attrs["type"],
                            src_name,
                            family,
                            bool(param in reqd_params),
                        )
                        for param, p_attrs in q_source.params.items()
                    },
                    table=q_source.params.get("table", {}),
                )

    def get_query_settings(self, family: str, query: str) -> QuerySource:
        """
        Get the QuerySource for the named `family` and `query`.

        Parameters
        ----------
        family : str
            Data family name
        query : str
            Query name

        Returns
        -------
        QuerySource
            Query settings object

        Raises
        ------
        KeyError
            If `family`.`query` could not be found.

        """
        q_source = self._provider.query_store.data_families.get(family, {}).get(query)
        if not q_source:
            raise KeyError(f"No query found for {family}.{query}")
        return q_source

    def get_query_pivot_settings(self, family: str, query: str) -> PivQuerySettings:
        """
        Get Pivot settings metadata for a query.

        Parameters
        ----------
        family : str
            Data family
        query : str
            Query name

        Returns
        -------
        PivQuerySettings
            Named tuple:

            - short_name - short name for the query
            - direct_func_entities - the entities to add a top level function to
            - assigned_entities - entities to assign the query to (if parameter
              mapping is not applicable).

        """
        qs_pivot = self.get_query_settings(family, query).metadata.get("pivot", {})
        return PivQuerySettings(
            short_name=qs_pivot.get("short_name"),
            direct_func_entities=qs_pivot.get("direct_func_entities"),
            assigned_entities=qs_pivot.get("assigned_entities"),
        )

    def get_queries_and_types_for_param(
        self, param: str
    ) -> Iterable[Tuple[str, str, str, Callable[[Any], Any]]]:
        """
        Get queries and parameter data types for `param`.

        Parameters
        ----------
        param : str
            The parameter name.

        Returns
        -------
        Iterable[Tuple[str, str, Callable[[Any], Any]]]
            Iterable of tuples listing:
            query_name, param_type, query_func

        """
        param_usage = self.param_usage.get(param)
        if param_usage is None:
            return []
        get_param_props = (
            (param.query, param.family, param.type, f"{param.family}.{param.query}")
            for param in param_usage
        )
        return [
            (q_name, q_family, p_type, getattr(self._provider, q_func), q_func)  # type: ignore
            for q_name, q_family, p_type, q_func in get_param_props
        ]

    def get_queries_for_param(
        self, param: str
    ) -> Iterable[Tuple[str, str, Callable[[Any], Any]]]:
        """
        Get the list of queries for a parameter.

        Parameters
        ----------
        param : str
            Parameter name

        Returns
        -------
        Iterable[Tuple[str, str, Callable[[Any], Any]]]
            Iterable of tuples listing:
            query_name, query_func

        """
        param_usage = self.param_usage.get(param)
        if not param_usage:
            return []
        return [
            (query_name, query_family, getattr(self._provider, query_func_name))
            for query_name, query_family, query_func_name in (
                (param.query, param.family, f"{param.family}.{param.query}")
                for param in self.param_usage.get(param)  # type:ignore
            )
        ]

    def get_params(self, query_func_name: str) -> Optional[QueryParams]:
        """
        Get the parameters for a query function.

        Parameters
        ----------
        query_func_name : str
            Query name - the name must be fully-qualified
            (e.g. 'WindowsSecurity.list_processes')

        Returns
        -------
        QueryParams
            QueryParams named tuple
            (all, required, full_required, param_attrs, table)

        """
        return self.query_params.get(query_func_name)

    def get_param_attrs(self, param_name: str) -> List[ParamAttrs]:
        """
        Get the attributes for a parameter name.

        Parameters
        ----------
        param_name : str
            Parameter name

        Returns
        -------
        List[ParamAttrs]
            List of ParamAttrs named tuples:
            (type, query, family, required)

        Notes
        -----
        Since parameters may be defined for multiple queries, the
        set of parameter attributes will be returned for each query.

        """
        return self.param_usage.get(param_name, [])


# Map of query parameter names to entities and the entity attrib
# corresponding to the query parameter value
PARAM_ENTITY_MAP: Dict[str, List[Tuple[Type[entities.Entity], str]]] = {
    "account_name": [(entities.Account, "Name")],
    "host_name": [(entities.Host, "fqdn")],
    "process_name": [(entities.Process, "ProcessFilePath")],
    "ip_address": [(entities.IpAddress, "Address")],
    "domain": [(entities.Dns, "DomainName")],
    "logon_session_id": [
        (entities.Process, "LogonSession"),
        (entities.HostLogonSession, "SessionId"),
        (entities.Account, "LogonId"),
    ],
    "proc_op": [],
    "process_id": [(entities.Process, "ProcessId")],
    "commandline": [(entities.Process, "CommandLine")],
    "url": [(entities.Url, "Url")],
    "file_hash": [(entities.File, "file_hash")],
    "resource_id": [(entities.AzureResource, "ResourceId")],
}

# aliases for parameters
PARAM_ENTITY_MAP["ip_address_list"] = PARAM_ENTITY_MAP["ip_address"]
PARAM_ENTITY_MAP["source_ip_list"] = PARAM_ENTITY_MAP["ip_address"]
PARAM_ENTITY_MAP["user"] = PARAM_ENTITY_MAP["account_name"]
PARAM_ENTITY_MAP["file_hash_list"] = PARAM_ENTITY_MAP["file_hash"]
PARAM_ENTITY_MAP["domain_list"] = PARAM_ENTITY_MAP["domain"]
PARAM_ENTITY_MAP["url_list"] = PARAM_ENTITY_MAP["url"]


def add_data_queries_to_entities(
    provider: QueryProvider, get_timespan: Callable[[], TimeSpan]
):
    """
    Add data queries from `provider` to entities.

    Parameters
    ----------
    provider : QueryProvider
        Query provider
    get_timespan : Callable[[], TimeSpan]
        Callback to get time span

    """
    q_funcs = PivotQueryFunctions(provider)

    add_queries_to_entities(
        prov_qry_funcs=q_funcs,
        container=provider.environment,
        get_timespan=get_timespan,
    )


# pylint: disable=too-many-locals


def add_queries_to_entities(
    prov_qry_funcs: PivotQueryFunctions,
    container: str,
    get_timespan: Callable[[], TimeSpan],
):
    """
    Add data queries to entities.

    Parameters
    ----------
    prov_qry_funcs : PivotQueryFunctions
        Collection of wrapped query functions
    container : str
        The name of the container to add query functions to
    get_timespan : Callable[[], TimeSpan]
        Function to get the current timespan.

    """
    # For each parameter in the parameter map
    for param_name, entity_list in PARAM_ENTITY_MAP.items():

        param_funcs = list(prov_qry_funcs.get_queries_for_param(param_name))
        if not (entity_list and param_funcs):
            continue
        entity_funcs = itertools.product(entity_list, param_funcs)
        # For each entity/query combo that uses this parameter
        for (entity_cls, _), (name, family, func) in entity_funcs:
            func_params = prov_qry_funcs.get_params(f"{family}.{name}")
            if not func_params or len(func_params.all) == 0:
                # ignore any functions without parameters
                continue

            # If multiple params - get the ones that are available in the same entity
            # We could in the future get parameters for connected (graph) entities.
            param_entities = {
                param: (ent, attr)
                for param, ent_list in PARAM_ENTITY_MAP.items()
                for ent, attr in ent_list
                if param in func_params.all and ent == entity_cls
            }
            # Build the map of param names to entity attributes
            attr_map = {
                param: ent_attr for param, (_, ent_attr) in param_entities.items()
            }
            # Wrap the function
            cls_func = _create_pivot_func(
                func, func_params.param_attrs, attr_map, get_timespan
            )
            # add a properties dict to the function
            setattr(
                cls_func,
                "pivot_properties",
                _create_piv_properties(name, param_entities, container),
            )
            q_piv_settings = prov_qry_funcs.get_query_pivot_settings(family, name)
            func_name = _format_func_name(name, func_params, q_piv_settings)

            # Add the wrapped function to the entity container
            query_container = getattr(entity_cls, container, None)
            if not query_container:
                query_container = QueryContainer()
                setattr(entity_cls, container, query_container)
            setattr(query_container, func_name, cls_func)

            # Also set this as a direct entity method if this entity is listed
            # in the query pivot "direct_func_entities" list
            if (
                q_piv_settings.direct_func_entities
                and entity_cls.__name__ in q_piv_settings.direct_func_entities
            ):
                dir_func_name = f"qry_{func_name}"
                setattr(entity_cls, dir_func_name, cls_func)


# pylint: enable=too-many-locals


def _format_func_name(name, func_params, q_piv_settings):
    # To help disambiguation we prefix the function name with
    # the table name (or short version)
    table_name = func_params.table.get("default", "")
    t_prefix = _TABLE_SHORTNAMES.get(table_name, table_name)
    if t_prefix and not t_prefix.endswith("_"):
        t_prefix = f"{t_prefix}_"
    # if query func has a short name, use that
    q_name = q_piv_settings.short_name or name
    return f"{t_prefix}{q_name}" if table_name else q_name


def _create_piv_properties(name, param_entities, container):
    return {
        "src_func_name": name,
        "src_class": "QueryProvider",
        "src_module": "msticpy.data.dataproviders",
        "entity_map": dict(param_entities.values()),
        "entity_container_name": container,
    }


def _create_pivot_func(
    func: Callable[[Any], pd.DataFrame],
    func_params: Dict[str, ParamAttrs],
    param_attrib_map: Dict[str, str],
    get_timespan: Callable[[], TimeSpan],
):
    """
    Wrap query function in to handle input parameters.

    Parameters
    ----------
    func : Callable[[Any], pd.DataFrame]
        The function to be wrapped
    func_params : Dict[str, ParamAttrs]
        Dict of parameters used by `func`
    param_attrib_map : Dict[str, str]
        Map of parameter name to entity attribute name.
    get_timespan : Callable[[], TimeSpan]
        The function to get the default timespan to use for queries.

    Returns
    -------
    Callable[Any, pd.DataFrame]
        The wrapped query function.

    Notes
    -----
    This function wraps the input `func` in two decorators.
    The first of these (create_data_func_exec) will execute
    the function once or many times (concat'ing the results) if
    the inputs are multi-valued.

    The second (in `wrapped_query_func`) handles the function being called
    wth arg[0] as an entity instance. In this case, it extracts the entity
    attributes (mapped to query function parameters) and uses those values
    as the input parameters to the function.

    """
    # initially wrap the function in a wrapper that actually does
    # the call to the query function.
    exec_query_func = _create_data_func_exec(func, func_params)

    # The outer wrapper handles instantiating query parameters at runtime
    @wraps(func)
    def wrapped_query_func(*args, **kwargs):
        """Wrap function to extract and map parameters."""
        # If time params not specified, add the global ones.
        time_params = {
            "start": kwargs.pop("start", get_timespan().start),
            "end": kwargs.pop("end", get_timespan().end),
        }
        # If this is an entity assume it's called as an instance method
        if args and isinstance(args[0], entities.Entity):
            value = args[0]
            param_dict = {
                param: getattr(value, attrib, None)
                for param, attrib in param_attrib_map.items()
                if hasattr(value, attrib)
            }
            return exec_query_func(**param_dict, **time_params, **kwargs)
        return exec_query_func(**time_params, **kwargs)

    return wrapped_query_func


def _create_data_func_exec(
    func: Callable[[Any], pd.DataFrame], func_params: Dict[str, ParamAttrs]
) -> Callable[[Any], pd.DataFrame]:
    """
    Wrap func to issue single or multiple calls to query.

    Parameters
    ----------
    func : Callable[[Any], pd.DataFrame]
        Query function to wrap
    func_params : Dict[str, ParamAttrs]
        Dictionary of function parameter definitions
        for this function.

    Returns
    -------
    Callable[[Any], pd.DataFrame]
        wrapped function.

    Notes
    -----
    Query functions have parameters that can accept single values
    or lists. This wrapper handles these cases depending on the input
    parameters.

    If the input is multi-values (i.e. a DataFrame or
    other iterable) and the required parameters accept list values,
    a single call to the query function is made. If any of the
    required parameters accept only single values, repeated calls
    are made to the query function and the results concatenated into
    a single DataFrame output.

    If the inputs are all single values, a single call is made, as normal.

    """

    @wraps(func)
    def call_data_query(**kwargs):
        """Call function handling input params in different formats."""
        func_kwargs = kwargs.copy()

        # The input is a DataFrame
        if "data" in kwargs:
            # If the input is a DF, we might be required to join
            join_type, left_on, right_on, j_ignore_case = get_join_params(func_kwargs)
            src_data = kwargs["data"] if join_type else None
            # Get the results of the query
            result_df = _exec_query_for_df(func, func_kwargs, func_params, kwargs)
            if join_type and isinstance(src_data, pd.DataFrame):
                if left_on and right_on:
                    # If explicit join keys
                    return join_result(
                        input_df=src_data,
                        result_df=result_df,
                        left_on=left_on,
                        right_on=right_on,
                        how=join_type,
                        ignore_case=j_ignore_case,
                    ).drop(columns="src_row_index", errors="ignore")
                if "src_row_index" in result_df.columns:
                    # Otherwise merge on index of source
                    return src_data.merge(
                        result_df,
                        left_index=True,
                        right_on="src_row_index",
                        how=join_type,
                    ).drop(columns="src_row_index", errors="ignore")

                warnings.warn(
                    "Cannot do an index merge on this result set. "
                    + "Please use an explicit column join using 'left_on' "
                    + "and 'right_on' join columns."
                )
            return result_df.drop(columns="src_row_index", errors="ignore")
        # The inputs are some mix of simple values and/or iterables.
        return _exec_query_for_values(func, func_kwargs, func_params, kwargs)

    return call_data_query  # type: ignore


def _exec_query_for_df(func, func_kwargs, func_params, parent_kwargs):
    """Execute `func` for DataFrame inputs."""
    src_df = func_kwargs.pop("data")
    parent_kwargs.pop("data")
    df_iter_params, list_params = _check_df_params_require_iter(
        func_params,
        src_df,
        func_kwargs,
        **parent_kwargs,
    )

    if not df_iter_params or df_iter_params.keys() == list_params.keys():
        # If there are no iter params that are not in the list_params
        # dict - we're only using list params - we're good to go
        return func(**list_params, **func_kwargs)

    # Even if we have list params, we can't use both list params and per-row
    # iteration so ignore these and run queries per row
    row_results = []
    # extact the DF subset of df_iter_params columns and iterate over each row
    for row_index, row in src_df[list(df_iter_params.values())].iterrows():
        # build a single-line dict of {param1: row_value1...}
        col_param_dict = {param: row[col] for param, col in df_iter_params.items()}
        # execute the function for each input row with key-value params from
        # col-name, col-value supplied as kwargs (along with any other kwargs)
        row_res_def = func(**col_param_dict, **func_kwargs)
        row_res_def["src_row_index"] = row_index
        row_results.append(row_res_def)
    return pd.concat(row_results, ignore_index=True)


def _check_df_params_require_iter(
    func_params: Dict[str, ParamAttrs],
    src_df: pd.DataFrame,
    func_kwargs: Dict[str, Any],
    **kwargs,
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Return params that require iteration and those that don't."""
    list_params: Dict[str, Any] = {}
    df_iter_params: Dict[str, Any] = {}
    for kw_name, arg in kwargs.items():
        if kw_name in _DEF_IGNORE_PARAM:
            continue
        if (
            arg not in src_df.columns
            or not isinstance(func_kwargs.get(kw_name), str)
            or kw_name not in func_params
        ):
            # Not intended/usable as a column specification
            continue
        col_name = func_kwargs.pop(kw_name)
        if func_params[kw_name].type == "list":
            # If the parameter accepts iterable types try to use the
            # values of that column directly
            list_params[kw_name] = list(src_df[col_name].values)
            # But also store it as a param that we might need to iterate through
        df_iter_params[kw_name] = col_name
    return df_iter_params, list_params


def _exec_query_for_values(func, func_kwargs, func_params, parent_kwargs):
    """Execute `func` for value (scalar or iterable) inputs."""
    var_iter_params, simple_params = _check_var_params_require_iter(
        func_params, func_kwargs, **parent_kwargs
    )

    if not var_iter_params or var_iter_params.keys() == simple_params.keys():
        # If there are no iter params that are not in the simple_params
        # dict - we're only using list params - we're good to go
        return func(**simple_params, **func_kwargs)

    for param in var_iter_params:
        simple_params.pop(param, None)

    # Even if we have list params, we can't use both list params and per-row
    # iteration so ignore these and run queries per row
    row_results = []
    # zip the value lists into tuples
    for row in zip(*(var_iter_params.values())):
        # build a single-line dict of {param1: row_value1...}
        col_param_dict = {param: row[idx] for idx, param in enumerate(var_iter_params)}
        row_results.append(func(**simple_params, **col_param_dict, **func_kwargs))
    return pd.concat(row_results, ignore_index=True)


def _check_var_params_require_iter(
    func_params: Dict[str, ParamAttrs], func_kwargs: Dict[str, Any], **kwargs
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Return params that require iteration and don't."""
    simple_params: Dict[str, Any] = {}
    var_iter_params: Dict[str, Any] = {}
    for kw_name, arg in kwargs.items():
        if kw_name in _DEF_IGNORE_PARAM:
            continue
        func_kwargs.pop(kw_name)
        if isinstance(arg, str) or not isinstance(arg, abc.Iterable):
            # treat as scalar/simple type
            simple_params[kw_name] = arg
        else:
            if func_params[kw_name].type == "list":
                # If the parameter accepts iterable types try to use the
                # values of that column directly
                simple_params[kw_name] = list(arg)
            # but also add it to the list of iterable params
            var_iter_params[kw_name] = list(arg)
    return var_iter_params, simple_params



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Entity sub-package."""
import difflib

from .account import Account
from .alert import Alert
from .azure_resource import AzureResource
from .cloud_application import CloudApplication
from .dns import Dns
from .entity import Entity
from .entity_enums import (  # noqa: F401
    Algorithm,
    ElevationToken,
    OSFamily,
    RegistryHive,
)
from .file import File
from .file_hash import FileHash
from .geo_location import GeoLocation
from .host import Host
from .host_logon_session import HostLogonSession
from .iot_device import IoTDevice
from .ip_address import IpAddress
from .mailbox import Mailbox
from .mail_cluster import MailCluster
from .mail_message import MailMessage
from .malware import Malware
from .network_connection import NetworkConnection
from .process import Process
from .registry_key import RegistryKey
from .registry_value import RegistryValue
from .security_group import SecurityGroup
from .submission_mail import SubmissionMail
from .threat_intelligence import Threatintelligence
from .unknown_entity import UnknownEntity
from .url import Url
from .cloud_logon_session import CloudLogonSession
from ..soc.incident import Incident

# Dictionary to map text names of types to the class.
Entity.ENTITY_NAME_MAP.update(
    {
        "account": Account,
        "azureresource": AzureResource,
        "azure-resource": AzureResource,
        "host": Host,
        "process": Process,
        "file": File,
        "cloudapplication": CloudApplication,
        "cloud-application": CloudApplication,
        "dns": Dns,
        "dnsresolve": Dns,
        "ipaddress": IpAddress,
        "iotdevice": IoTDevice,
        "ip": IpAddress,
        "networkconnection": NetworkConnection,
        "mailbox": Mailbox,
        "mail-message": MailMessage,
        "mailmessage": MailMessage,
        "mail-cluster": MailCluster,
        "mailcluster": MailCluster,
        "malware": Malware,
        "registry-key": RegistryKey,
        "registrykey": RegistryKey,
        "registry-value": RegistryValue,
        "registryvalue": RegistryValue,
        "host-logon-session": HostLogonSession,
        "hostlogonsession": HostLogonSession,
        "filehash": FileHash,
        "security-group": SecurityGroup,
        "securitygroup": SecurityGroup,
        "SubmissionMail": SubmissionMail,
        "alerts": Alert,
        "alert": Alert,
        "threatintelligence": Threatintelligence,
        "url": Url,
        "unknown": UnknownEntity,
        "geolocation": GeoLocation,
        "location": GeoLocation,
        "incident": Incident,
        "cloud-logon-session": CloudLogonSession,
    }
)


def find_entity(entity):
    """Find entity name."""
    entity_cf = entity.casefold()
    entity_classes = {
        cls.__name__.casefold(): cls for cls in Entity.ENTITY_NAME_MAP.values()
    }
    if entity_cf in Entity.ENTITY_NAME_MAP:
        print(f"Match found '{Entity.ENTITY_NAME_MAP[entity].__name__}'")
        return Entity.ENTITY_NAME_MAP[entity]
    if entity_cf in entity_classes:
        print(f"Match found '{entity_classes[entity_cf].__name__}'")
        return entity_classes[entity_cf]
    # Try to find the closest matches
    closest = difflib.get_close_matches(entity, entity_classes.keys(), cutoff=0.4)
    mssg = [f"No exact match found for '{entity}'. "]
    if len(closest) == 1:
        mssg.append(f"Closest match is '{entity_classes[closest[0]].__name__}'")
    elif closest:
        match_list = [f"'{entity_classes[mtch].__name__}'" for mtch in closest]
        mssg.append(f"Closest matches are {', '.join(match_list)}")
    else:
        mssg.extend(
            [
                "No close match found. Entities available:",
                *(cls.__name__ for cls in entity_classes.values()),
            ]
        )
    print("\n".join(mssg))
    return None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MSTICPY Config modules."""

# flake8: noqa: F403
from .mp_config_edit import MpConfigEdit
from .mp_config_file import MpConfigFile
from .mp_config_control import MpConfigControls



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Pivot TI Provider helper functions."""
from collections import defaultdict
from typing import Callable, Dict, Optional, Set, Tuple, Type

import pandas as pd

from .._version import VERSION
from ..data.query_container import QueryContainer
from ..sectools.tilookup import TILookup
from ..sectools.tiproviders.ti_provider_base import TIPivotProvider
from . import entities
from .pivot_register import PivotRegistration, create_pivot_func

__version__ = VERSION
__author__ = "Ian Hellen"

IOC_TYPES = {"ipv4", "ipv6", "dns", "file_hash", "url"}

TI_ENTITY_ATTRIBS: Dict[str, Tuple[Type, str]] = {
    "ipv4": (entities.IpAddress, "Address"),
    "ipv6": (entities.IpAddress, "Address"),
    "ip": (entities.IpAddress, "Address"),
    "dns": (entities.Dns, "DomainName"),
    "file_hash": (entities.File, "file_hash"),
    "file_path": (entities.File, "file_hash"),
    "url": (entities.Url, "Url"),
}


def add_ioc_queries_to_entities(ti_lookup: TILookup, container: str = "ti", **kwargs):
    """
    Add TI functions to entities.

    Parameters
    ----------
    ti_lookup : TILookup
        TILookup instance.
    container : str
        The name of the container to add query functions to

    """
    ioc_queries = create_ti_pivot_funcs(ti_lookup)
    for ioc, ioc_funcs in ioc_queries.items():
        if "debug" in kwargs:
            print(ioc, ioc_funcs)
        entity, _ = TI_ENTITY_ATTRIBS[ioc]
        if entity:
            for f_name, func in ioc_funcs.items():
                if "debug" in kwargs:
                    print(ioc, f_name, func)
                query_container = getattr(entity, container, None)
                if not query_container:
                    query_container = QueryContainer()
                    setattr(entity, container, query_container)
                setattr(query_container, f_name, func)

                # Create shortcuts for non-provider-specific funcs
                if f_name.endswith(ioc):
                    short_func_name = f"ti{f_name}"
                    setattr(entity, short_func_name, func)


def create_ti_pivot_funcs(ti_lookup: TILookup):
    """Create the TI Pivot functions."""
    ioc_type_supp = _get_supported_ioc_types(ti_lookup)
    ioc_queries: Dict[str, Dict[str, Callable[..., pd.DataFrame]]] = defaultdict(dict)

    # Add functions for ioc types that will call all providers
    # Non-IP types
    ioc_queries.update(_get_non_ip_functions(ioc_type_supp, ti_lookup))
    # Special case for ipv4 and ipv6 - we want to merge these into "ip" if these are equivalent
    ioc_queries.update(_get_ip_functions(ioc_type_supp, ti_lookup))

    # Add functions for provider-specific lookup function names
    # These have a "_provider_name" suffix
    for prov, ioc_set in ioc_type_supp.items():
        for ioc in ioc_set:
            ioc_name = _merged_ip_ioc_type(ioc, ti_lookup.loaded_providers.get(prov))
            if not ioc_name:
                continue
            _, func_name, func = _create_lookup_func(ti_lookup, ioc, ioc, [prov])
            func_name = f"{func_name}_{prov}"
            ioc_queries[ioc][func_name] = func
    return ioc_queries


def register_ti_pivot_providers(ti_lookup: TILookup, pivot: "Pivot"):  # type: ignore # noqa: F821
    """Register pivot functions from TI providers."""
    for _, ti_prov in ti_lookup.loaded_providers.items():
        if isinstance(ti_prov, TIPivotProvider):
            ti_prov.register_pivots(PivotRegistration, pivot)


def _get_supported_ioc_types(ti_lookup: TILookup) -> Dict[str, Set[str]]:
    return {
        ti_prov_name: set(ti_prov.supported_types) & IOC_TYPES
        for ti_prov_name, ti_prov in ti_lookup.loaded_providers.items()
    }


def _prov_ipv4v6_equal(ti_provider) -> bool:
    if not ti_provider:
        return False
    ti_queries = ti_provider.ioc_query_defs
    return ti_queries.get("ipv4") is not None and (
        ti_queries.get("ipv6") is None
        or ti_queries.get("ipv4") == ti_queries.get("ipv6")
    )


def _merged_ip_ioc_type(ioc, ti_provider) -> Optional[str]:
    if ioc == "ipv4" and _prov_ipv4v6_equal(ti_provider):
        return "ip"
    if ioc == "ipv6" and _prov_ipv4v6_equal(ti_provider):
        return None
    return ioc


def _create_lookup_func(
    ti_lookup: TILookup, ioc, ioc_name, providers
) -> Tuple[str, str, Callable[..., pd.DataFrame]]:

    suffix = f"_{ioc_name}"
    short_func_name = f"lookup{suffix}"
    func_name = f"{short_func_name}_{ioc_name}"
    params_dict = {"providers": providers, "ioc_type": ioc}

    entity_cls, entity_attr = TI_ENTITY_ATTRIBS[ioc]

    pivot_reg = PivotRegistration(
        src_func_name=ti_lookup.lookup_iocs.__name__,
        input_type="dataframe",
        entity_map={entity_cls.__name__: entity_attr},
        func_df_param_name="data",
        func_df_col_param_name="obs_col",
        func_out_column_name="Ioc",
        func_static_params=params_dict,
    )
    return (
        func_name,
        short_func_name,
        create_pivot_func(target_func=ti_lookup.lookup_iocs, pivot_reg=pivot_reg),
    )


def _get_non_ip_functions(ioc_type_supp, ti_lookup):
    """Get functions for non-IP IoC types."""
    ioc_queries = defaultdict(dict)
    for ioc in IOC_TYPES - {"ipv4", "ipv6"}:
        supporting_provs = [
            prov for prov, supp_types in ioc_type_supp.items() if ioc in supp_types
        ]
        _, func_name, func = _create_lookup_func(ti_lookup, ioc, ioc, supporting_provs)
        ioc_queries[ioc][func_name] = func
    return ioc_queries


def _get_ip_functions(ioc_type_supp, ti_lookup):
    """Get functions for IP IoC Types."""
    # Special case for ipv4 and ipv6
    # we want to merge these into "ip" if these are equivalent
    ioc_queries = defaultdict(dict)
    # Special case for ipv4 and ipv6 - we want to merge these into "ip" if these are equivalent
    ip_types = {"ipv4", "ipv6"}
    ip_all_provs = [
        prov for prov, supp_types in ioc_type_supp.items() if ip_types & supp_types
    ]

    # Register providers where IP v4 and v6 are equivalent, or only support ipv4, as "ip"
    ip_gen_provs = [
        prov
        for prov in ip_all_provs
        if _prov_ipv4v6_equal(ti_lookup.loaded_providers.get(prov))
        or "ipv6" not in ioc_type_supp[prov]
    ]

    _, func_name, func = _create_lookup_func(ti_lookup, "ipv4", "ip", ip_gen_provs)
    ioc_queries["ip"][func_name] = func

    # Also register IP v4 and v6 specific queries
    ip_v4_provs = [
        prov
        for prov in ip_all_provs
        if "ipv4" in ioc_type_supp[prov]
        if prov not in ip_gen_provs
    ]
    _, func_name, func = _create_lookup_func(ti_lookup, "ipv4", "ipv4", ip_v4_provs)
    ioc_queries["ipv4"][func_name] = func

    ip_v6_provs = [
        prov
        for prov in ip_all_provs
        if "ipv6" in ioc_type_supp[prov]
        if prov not in ip_gen_provs
    ]
    _, func_name, func = _create_lookup_func(ti_lookup, "ipv6", "ipv6", ip_v6_provs)
    ioc_queries["ipv6"][func_name] = func

    return ioc_queries



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Keyvault client settings."""

import warnings
from typing import Any, List, Optional

from .._version import VERSION
from . import pkg_config as config
from .azure_auth_core import AzureCloudConfig
from .exceptions import MsticpyKeyVaultConfigError
from .utility import export
from .cloud_mappings import create_cloud_ep_dict, create_cloud_suf_dict

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class KeyVaultSettings:
    """
    KeyVaultSettings class - reads settings from msticpyconfig.

    Notes
    -----
    The KeyVault section in msticpyconfig.yaml can contain
    the following::

        KeyVault:
            TenantId: {tenantid-to-use-for-authentication}
            SubscriptionId: {subscriptionid-containing-vault}
            ResourceGroup: {resource-group-containing-vault}
            AzureRegion: {region-for-vault}
            VaultName: {vault-name}
            UseKeyring: True
            Authority: global

    `SubscriptionId`, `ResourceGroup` and `AzureRegion` are only
    used when creating new vaults.
    `UseKeyring` instructs the `SecretsClient` to cache Keyvault
    secrets locally using Python keyring.
    `Authority` is one of 'global', 'usgov', 'de', 'cn'
    Alternatively, you can specify `AuthorityURI` with the value
    pointing to the URI for logon requests.

    """

    AAD_AUTHORITIES = create_cloud_ep_dict("active_directory")
    RES_MGMT_URIS = create_cloud_ep_dict("resource_manager")
    KV_SUFFIXES = create_cloud_suf_dict("keyvault_dns")
    KV_URIS = {
        cloud: f"https://{{vault}}{suffix}" for cloud, suffix in KV_SUFFIXES.items()
    }

    # Azure CLI Client ID
    CLIENT_ID = "04b07795-8ddb-461a-bbee-02f9e1bf7b46"  # xplat

    def __init__(self):
        """
        Initialize new instance of KeyVault Settings.

        Raises
        ------
        MsticpyKeyVaultConfigError
            If no Key Vault settings are found in
            msticpyconfig.yaml.

        """
        self.authority: Optional[str] = None
        self.auth_methods: List[str] = []
        try:
            kv_config = config.get_config("KeyVault")
        except KeyError as err:
            raise MsticpyKeyVaultConfigError(
                "No KeyVault section found in msticpyconfig.yaml",
                title="missing Key Vault configuration",
            ) from err
        norm_settings = {key.casefold(): val for key, val in kv_config.items()}
        self.__dict__.update(norm_settings)

        self._get_auth_methods_from_settings()
        self._get_authority_from_settings()

    def _get_auth_methods_from_settings(self):
        """Retrieve authentication methods from settings."""
        self.auth_methods = AzureCloudConfig().auth_methods

    def _get_authority_from_settings(self):
        """Get the authority (AAD) URI from settings."""
        if "authorityuri" in self:
            # For BlueHound compat - the "authority_uri" can be set directly
            # as a property of the object
            rev_lookup = {uri.casefold(): code for code, uri in self.AAD_AUTHORITIES}
            self.authority = rev_lookup.get(
                self["authorityuri"].casefold(), "global"
            ).casefold()
        elif not self.authority:
            self.authority = AzureCloudConfig().cloud

    def __getitem__(self, key: str):
        """Allow property get using dictionary key syntax."""
        if key.casefold() in self.__dict__:
            return self.__dict__[key.casefold()]
        raise KeyError

    def __setitem__(self, key: str, value: Any):
        """Allow property set using dictionary key syntax."""
        self.__dict__[key.casefold()] = value

    def __contains__(self, key: str):
        """Return true if key is a valid attribute."""
        return key.casefold() in self.__dict__

    def get(self, key: str, default: Any = None) -> Any:
        """Return dict value or default."""
        return self.__dict__.get(key.casefold(), default)

    @property
    def cloud(self) -> str:
        """Return the cloud for the settings."""
        return self.get("authority", "global").casefold()

    @property
    def authority_uri(self) -> str:
        """
        Return authority URI for cloud.

        Returns
        -------
        str
            Authority URI

        """
        if "authorityuri" in self:
            return self["authorityuri"]
        if self.cloud in self.AAD_AUTHORITIES:
            return self.AAD_AUTHORITIES[self.cloud]
        return self.AAD_AUTHORITIES["global"]

    @property
    def keyvault_uri(self) -> Optional[str]:
        """Return KeyVault URI template for current cloud."""
        kv_uri = self.KV_URIS.get(self.cloud)
        if not kv_uri:
            mssg = f"Could not find a valid KeyVault endpoint for {self.cloud}"
            warnings.warn(mssg)
        return kv_uri

    @property
    def mgmt_uri(self) -> Optional[str]:
        """Return Azure management URI template for current cloud."""
        mgmt_uri = self.RES_MGMT_URIS.get(self.cloud)
        if not mgmt_uri:
            mssg = f"Could not find a valid KeyVault endpoint for {self.cloud}"
            warnings.warn(mssg)
        return mgmt_uri

    def get_tenant_authority_uri(
        self, authority_uri: str = None, tenant: str = None
    ) -> str:
        """
        Return authority URI for tenant.

        Parameters
        ----------
        authority_uri : str, optional
            The authority URI - otherwise try to retrieve
            from settings
        tenant : str, optional
            TenantID or name, by default None.
            If not passed as a parameter try to get tenant from
            KeyVault configuration in msticpyconfig.yaml

        Returns
        -------
        str
            Tenant Authority

        Raises
        ------
        KeyVaultConfigException
            If tenant is not defined.

        """
        auth = authority_uri or self.authority_uri.strip()
        if not tenant:
            tenant = self.get("tenantid")
        if not tenant:
            raise MsticpyKeyVaultConfigError(
                "Could not get TenantId from function parameters or configuration.",
                "Please add this to the KeyVault section of msticpyconfig.yaml",
                title="missing tenant ID value.",
            )
        if auth.endswith("/"):
            return auth + tenant.strip()
        return auth + "/" + tenant.strip()

    def get_tenant_authority_host(
        self, authority_uri: str = None, tenant: str = None
    ) -> str:
        """
        Return tenant authority URI with no leading scheme.

        Parameters
        ----------
        authority_uri : str, optional
            The authority URI - otherwise try to retrieve
            from settings
        tenant : str, optional
            TenantID or name, by default None.
            If not passed as a parameter try to get tenant from
            KeyVault configuration in msticpyconfig.yaml

        Returns
        -------
        str
            Tenant Authority

        Raises
        ------
        KeyVaultConfigException
            If tenant is not defined.

        """
        if not tenant:
            tenant = self.get("tenantid")
        if not tenant:
            raise MsticpyKeyVaultConfigError(
                "Could not get TenantId from function parameters or configuration.",
                "Please add this to the KeyVault section of msticpyconfig.yaml",
                title="missing tenant ID value.",
            )
        return (
            self.get_tenant_authority_uri(authority_uri, tenant)
            .lower()
            .replace("https://", "")
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Process tree builder routines for MDE process data."""
from typing import Dict, Tuple, Union

import numpy as np
import pandas as pd

from .proc_tree_schema import (
    ProcSchema,
)
from ..data.query_defns import ensure_df_datetimes
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


_MDE_NON_STD_COL_MAP = {
    "InitiatingProcessImageCreationTimeUtc": "CreatedProcessFileCreationTime",
    "InitiatingProcessImageMd5": "CreatedProcessFileMd5",
    "InitiatingProcessImageFilePath": "CreatedProcessFilePath",
    "InitiatingProcessImageSha1": "CreatedProcessFileSha1",
    "InitiatingProcessImageSha256": "CreatedProcessFileSha256",
    "InitiatingProcessImageFileSizeInBytes": "CreatedProcessFileSize",
    "InitiatingProcessParentProcessId": "CreatedProcessParentId",
    "InitiatingProcessParentProcessName": "CreatedProcessParentName",
    "InitiatingProcessParentCreationTime": "CreatedProcessParentCreationTimeUtc",
}

_MDE_TIMESTAMP_COLS = [
    "CreatedProcessCreationTime",
    "InitiatingProcessCreationTime",
    "CreatedProcessParentCreationTimeUtc",
    "InitiatingProcessParentCreationTime",
]

TS_FMT_STRING = "%Y-%m-%d %H:%M:%S.%f"
PARENT_KEY = "parent_key"
PROC_KEY = "proc_key"


def extract_process_tree(data: pd.DataFrame, debug: bool = False) -> pd.DataFrame:
    """
    Build a process tree from raw MDE process logs.

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame of process events.
    debug: bool, optional
        Turn on additional debugging output, by default False.

    Returns
    -------
    pd.DataFrame
        Process tree DataFrame with child->parent keys and
        extracted parent processes from child data.

    """
    data = ensure_df_datetimes(data, columns=_MDE_TIMESTAMP_COLS)
    par_child_col_map = _get_par_child_col_mapping(data)
    inferred_parents = _extract_missing_parents(data, par_child_col_map, debug=debug)
    missing_par_uniq = _get_unique_parents(inferred_parents, debug)
    all_proc_par = pd.concat([data, missing_par_uniq], ignore_index=True)
    missing_gps = _extract_missing_gparents(all_proc_par)
    return pd.concat([all_proc_par, missing_gps], ignore_index=True, axis=0)


def _sort_df_by_time(data, column="CreatedProcessCreationTime"):
    """Return dataframe sorted by specified column and reset index."""
    return data.sort_values(column, ascending=True).reset_index().drop(columns="index")


def _add_proc_key(
    data: pd.DataFrame,
    key_name: str,
    proc_name_col: str,
    proc_id_col: str,
    timestamp_col: str,
):
    """
    Add process/parent key to data.

    Parameters
    ----------
    data : pd.DataFrame
        Input data
    key_name : str
        The name of the key to create
    proc_name_col : str
        Process name column
    proc_id_col : str
        Process ID column
    timestamp_col : str
        Process time stamp column.

    Notes
    -----
    This function adds the key column to the passed `data` dataframe.

    """
    data[key_name] = (
        data[proc_name_col].str.lower()
        + "|"
        + data[proc_id_col].astype(str).str.lower()
        + "|"
        + data[timestamp_col].dt.round("10us").dt.strftime(TS_FMT_STRING)
    )


def _extract_missing_parents(
    data: pd.DataFrame, col_mapping: Dict[str, str], debug: bool = False
) -> pd.DataFrame:
    """Return parent processes that are not in the created process set."""
    # save the source index
    data["src_index"] = data.index
    # add process key
    _add_proc_key(
        data,
        PROC_KEY,
        "CreatedProcessName",
        "CreatedProcessId",
        "CreatedProcessCreationTime",
    )

    # Create full process file path for initiating processes
    data["InitiatingProcessFileName"] = (
        data.InitiatingProcessImageFilePath + "\\" + data.InitiatingProcessName
    )
    # Create parent key
    _add_proc_key(
        data,
        PARENT_KEY,
        "InitiatingProcessName",
        "InitiatingProcessId",
        "InitiatingProcessCreationTime",
    )

    # Separate the InitiatingProcess columns from the DF
    parent_cols = set(data.filter(regex="Initiating.*").columns)
    non_par_cols = set(data.columns) - parent_cols
    non_par_cols.remove("src_index")
    # print(non_par_cols)

    # merge the original data with the parent rows
    merged_parents = data.filter(
        regex="Initiating.*|parent_key|src_index"
    ).merge(  # parents
        data.filter(non_par_cols),  # created_procs
        left_on=PARENT_KEY,
        right_on=PROC_KEY,
        suffixes=("_child", "_par"),
        how="left",
    )

    # Any NA results from the merge indicate that the parent process
    # is not itself in the list of processes
    missing_parents = merged_parents[merged_parents["CreatedProcessParentId"].isna()]
    missing_parents = (
        missing_parents.dropna(axis=1, how="all")
        .rename(columns=col_mapping)
        .rename(columns={"parent_key_child": PROC_KEY})
        .drop(columns=["InitiatingProcessFileName"])
    )
    missing_parents["CreatedProcessFilePath"] = (
        missing_parents.CreatedProcessFilePath
        + "\\"
        + missing_parents.CreatedProcessName
    )
    missing_parents = _sort_df_by_time(missing_parents)
    if debug:
        print("missing parent procs", len(missing_parents))
        # All successful merges indicate that the parent IS in the list of processes
        found_parents = merged_parents[~merged_parents["CreatedProcessParentId"].isna()]
        print("existing parent procs", len(found_parents))
        mpar_uniq_test = (
            missing_parents.drop(columns="src_index").groupby("proc_key").nunique()
        )
        if not mpar_uniq_test[mpar_uniq_test > 1].dropna(how="all").empty:
            print("Error - some extracted parents have duplicate keys")
            print(mpar_uniq_test[mpar_uniq_test > 1].dropna(how="all"))
    return missing_parents


# Get unique parents and add the parent key
def _get_unique_parents(data, debug=False):
    """Return unique set of processes."""
    if "src_index" in data.columns:
        data = data.drop(columns="src_index")
    missing_par_uniq = data.drop_duplicates().copy()
    missing_par_uniq["InitiatingProcessName"] = missing_par_uniq.apply(
        lambda row: row.CreatedProcessParentName.split("\\")[-1], axis=1
    )
    if debug:
        print("unique:", missing_par_uniq.shape, "original:", data.shape)
    _add_proc_key(
        missing_par_uniq,
        PARENT_KEY,
        "InitiatingProcessName",
        "CreatedProcessParentId",
        "CreatedProcessParentCreationTimeUtc",
    )
    return missing_par_uniq


def _split_file_path(
    input_path: str,
    path_col: str = "CreatedProcessFilePath",
    file_col: str = "CreatedProcessName",
    separator: str = "\\",
) -> Dict[str, Union[str, float]]:
    """Split file path in to folder/stem."""
    f_path: Union[str, float] = np.nan
    f_stem: Union[str, float] = np.nan
    try:
        f_path, _, f_stem = input_path.rpartition(separator)
    except AttributeError:
        pass
    return {path_col: f_path, file_col: f_stem}


def _extract_missing_gparents(data):
    """Return grandparent processes for any procs not in Createdprocesses."""
    missing_gps = (
        data[~data.parent_key.isin(data.proc_key)]
        .filter(regex=".*Parent.*")
        .drop_duplicates()
    )
    missing_gps_file_split = missing_gps.apply(
        lambda proc: _split_file_path(proc.CreatedProcessParentName),
        axis=1,
        result_type="expand",
    )
    missing_gps = (
        missing_gps.join(missing_gps_file_split)
        .drop(
            columns=[
                "InitiatingProcessParentCreationTime",
                "InitiatingProcessParentProcessId",
                "InitiatingProcessParentProcessName",
                "CreatedProcessFilePath",
            ]
        )
        .rename(
            columns={
                "CreatedProcessParentCreationTimeUtc": "CreatedProcessCreationTime",
                "CreatedProcessParentName": "CreatedProcessFilePath",
                "CreatedProcessParentId": "CreatedProcessId",
            }
        )
    )
    _add_proc_key(
        missing_gps,
        "proc_key",
        "CreatedProcessName",
        "CreatedProcessId",
        "CreatedProcessCreationTime",
    )
    return missing_gps


def _get_par_child_col_mapping(data: pd.DataFrame) -> Dict[str, str]:
    """Return a mapping between parent and child column names."""
    created_proc_cols = _remove_col_prefix(data, "Created")
    init_proc_cols = _remove_col_prefix(data, "Initiating")
    init_proc_col_mapping, _ = _map_columns(created_proc_cols, init_proc_cols)
    return {**init_proc_col_mapping, **_MDE_NON_STD_COL_MAP}


def _remove_col_prefix(data: pd.DataFrame, prefix: str) -> Dict[str, str]:
    """Return a mapping of column stems and columns with `prefix`."""
    return {
        col.replace(prefix, ""): col for col in data.columns if col.startswith(prefix)
    }


def _map_columns(
    created_cols: Dict[str, str], init_cols: Dict[str, str]
) -> Tuple[Dict[str, str], Dict[str, str]]:
    """Return Initiating -> Created column mapping."""
    col_mapping = {}
    unmapped = {}
    for col_stem, col in init_cols.items():
        if col_stem in created_cols:
            col_mapping[col] = created_cols[col_stem]
        else:
            unmapped[col_stem] = col
    return col_mapping, unmapped


_SENTINEL_MDE_MAP = {
    "AccountDomain": "CreatedProcessAccountDomainName",
    "AccountName": "CreatedProcessAccountName",
    "AccountSid": "CreatedProcessAccountSid",
    "DeviceId": "MachineId",
    "DeviceName": "ComputerDnsName",
    "FileName": "CreatedProcessName",
    "FolderPath": "CreatedProcessFilePath",
    "InitiatingProcessAccountDomain": "InitiatingProcessAccountDomainName",
    "InitiatingProcessCommandLine": "Process_CommandLine",
    "InitiatingProcessFileName": "InitiatingProcessName",
    "InitiatingProcessFolderPath": "InitiatingProcessImageFilePath",
    "InitiatingProcessMD5": "InitiatingProcessImageMd5",
    "InitiatingProcessParentFileName": "InitiatingProcessParentProcessName",
    "InitiatingProcessParentId": "InitiatingProcessParentProcessId",
    "InitiatingProcessSHA1": "InitiatingProcessImageSha1",
    "InitiatingProcessSHA256": "InitiatingProcessImageSha256",
    "InitiatingProcessTokenElevation": "InitiatingProcessTokenElevationType",
    "MD5": "CreatedProcessFileMd5",
    "ProcessCommandLine": "CreatedProcessCommandLine",
    "ProcessCreationTime": "CreatedProcessCreationTime",
    "ProcessId": "CreatedProcessId",
    "ProcessIntegrityLevel": "CreatedProcessIntegrityLevel",
    "ProcessTokenElevation": "CreatedProcessTokenElevationType",
    "SHA1": "CreatedProcessFileSha1",
    "SHA256": "CreatedProcessFileSha256",
}

_UNK_TIME = pd.Timestamp("1970-01-01", tz="UTC")


def convert_mde_schema_to_internal(
    data: pd.DataFrame, schema: ProcSchema
) -> pd.DataFrame:
    """
    Convert DeviceProcessEvents schema data to internal MDE schema.

    Parameters
    ----------
    data : pd.DataFrame
        Input data in MS Sentinel schema.
    schema : ProcSchema
        The mapping schema for the data set.

    Returns
    -------
    pd.DataFrame
        Reformatted data into MDE internal schema.

    """
    # Fill in missing timestamps with placeholder
    data["ProcessCreationTime"] = data[schema.time_stamp].fillna(_UNK_TIME)
    data["InitiatingProcessCreationTime"] = data.InitiatingProcessCreationTime.fillna(
        _UNK_TIME
    )
    data[
        "InitiatingProcessParentCreationTime"
    ] = data.InitiatingProcessParentCreationTime.fillna(_UNK_TIME)

    # Proc tree code references CreateProcessParentId
    # This should be the same as InitiatingProcessParentId
    data["CreatedProcessParentId"] = data[schema.parent_id]

    # Put a value in parent procs with no name
    null_proc_parent = data[schema.parent_name] == ""
    data.loc[null_proc_parent, schema.parent_name] = "unknown"

    # Extract InitiatingProc folder path - remove stem
    data["InitiatingProcessFolderPath"] = data.InitiatingProcessFolderPath.apply(
        lambda x: x.rsplit("\\", maxsplit=1)[0]
    )

    return data.rename(columns=_SENTINEL_MDE_MAP)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Helper functions for configuration settings."""
from collections import UserDict
import os
from typing import Any, Dict, Optional, Union, Callable
import warnings

import attr
from attr import Factory

from .._version import VERSION
from .exceptions import MsticpyImportExtraError
from . import pkg_config as config

try:
    from .secret_settings import SecretsClient

    _SECRETS_ENABLED = True
except ImportError:
    _SECRETS_ENABLED = False


__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-few-public-methods, too-many-ancestors
class ProviderArgs(UserDict):
    """ProviderArgs dictionary."""

    def __getitem__(self, key):
        """Return key value via SecretsClient.read_secret."""
        if key not in self.data:
            raise KeyError(key)
        if _SECRETS_ENABLED:
            return SecretsClient.read_secret(self.data[key])
        return self.data[key]


@attr.s(auto_attribs=True)
class ProviderSettings:
    """Provider settings."""

    name: str
    description: str
    provider: Optional[str] = None
    args: ProviderArgs = Factory(ProviderArgs)  # type: ignore
    primary: bool = False


# pylint: enable=too-few-public-methods, too-many-ancestors


_SECRETS_CLIENT: Any = None
if "KeyVault" in config.settings and config.settings["KeyVault"] and _SECRETS_ENABLED:
    _SECRETS_CLIENT = SecretsClient()


def get_provider_settings(config_section="TIProviders") -> Dict[str, ProviderSettings]:
    """
    Read Provider settings from package config.

    Parameters
    ----------
    config_section : str, optional
        [description], by default "TIProviders"

    Returns
    -------
    Dict[str, ProviderSettings]
        Provider settings indexed by provider name.

    """
    # pylint: disable=global-statement
    global _SECRETS_CLIENT
    # pylint: enable=global-statement
    if "KeyVault" in config.settings and config.settings["KeyVault"]:
        if _SECRETS_CLIENT is None and _SECRETS_ENABLED:
            print(
                "KeyVault enabled. Secrets access may require additional authentication."
            )
            _SECRETS_CLIENT = SecretsClient()
    else:
        _SECRETS_CLIENT = None
    section_settings = config.settings.get(config_section)
    if not section_settings:
        return {}

    settings = {}
    for provider, item_settings in section_settings.items():
        prov_args = item_settings.get("Args")
        prov_settings = ProviderSettings(
            name=provider,
            description=item_settings.get("Description"),
            args=_get_setting_args(
                config_section=config_section,
                provider_name=provider,
                prov_args=prov_args,
            ),
            primary=item_settings.get("Primary", False),
            provider=item_settings.get("Provider", provider),
        )
        settings[provider] = prov_settings

    return settings


def reload_settings():
    """
    Reload settings from config files.

    Parameters
    ----------
    clear_keyring : bool, optional
        Clears any secrets cached in keyring, by default False

    """
    config.refresh_config()


def _get_setting_args(
    config_section: str, provider_name: str, prov_args: Optional[Dict[str, Any]]
) -> ProviderArgs:
    """Extract the provider args from the settings."""
    if not prov_args:
        return ProviderArgs()
    name_map = {
        "workspaceid": "workspace_id",
        "tenantid": "tenant_id",
        "subscriptionid": "subscription_id",
    }
    return _get_settings(
        config_section=config_section,
        provider_name=provider_name,
        conf_group=prov_args,
        name_map=name_map,
    )


def _get_settings(
    config_section: str,
    provider_name: str,
    conf_group: Optional[Dict[str, Any]],
    name_map: Optional[Dict[str, str]] = None,
) -> ProviderArgs:
    """
    Lookup configuration values config, environment or KeyVault.

    Parameters
    ----------
    config_section : str
        Configuration section
    provider_name: str
        The name of the provider section
    conf_group : Optional[Dict[str, Any]]
        The configuration dictionary
    name_map : Optional[Dict[str, str]], optional
        Optional mapping to re-write setting names,
        by default None

    Returns
    -------
    ProviderArgs
        Dictionary of resolved settings

    Raises
    ------
    NotImplementedError
        Keyvault storage is not yet implemented

    """
    if not conf_group:
        return ProviderArgs()
    setting_dict: ProviderArgs = ProviderArgs(conf_group.copy())

    for arg_name, arg_value in conf_group.items():
        target_name = arg_name
        if name_map:
            target_name = name_map.get(target_name.casefold(), target_name)

        if isinstance(arg_value, str):
            setting_dict[target_name] = arg_value
        elif isinstance(arg_value, dict):
            try:
                setting_dict[target_name] = _fetch_setting(
                    config_section, provider_name, arg_name, arg_value
                )  # type: ignore
            except NotImplementedError:
                warnings.warn(
                    f"Setting type for setting {arg_value} not yet implemented. "
                )
    return setting_dict


def _fetch_setting(
    config_section: str,
    provider_name: str,
    arg_name: str,
    config_setting: Dict[str, Any],
) -> Union[Optional[str], Callable[[], Any]]:
    """Return required value for indirect settings (e.g. getting env var)."""
    if "EnvironmentVar" in config_setting:
        env_value = os.environ.get(config_setting["EnvironmentVar"])
        if not env_value:
            warnings.warn(
                f"Environment variable {config_setting['EnvironmentVar']}"
                + f" (provider {provider_name})"
                + " was not set"
            )
        return env_value
    if "KeyVault" in config_setting:
        if not _SECRETS_ENABLED:
            raise MsticpyImportExtraError(
                "Cannot use this feature without Key Vault support installed",
                title="Error importing Loading Key Vault and/or keyring libaries",
                extra="keyvault",
            )
        if not _SECRETS_CLIENT:
            warnings.warn(
                "Cannot use a KeyVault configuration setting without"
                + "a KeyVault configuration section in msticpyconfig.yaml"
                + f" (provider {provider_name})"
            )
            return None
        config_path = [config_section, provider_name, "Args", arg_name]
        return _SECRETS_CLIENT.get_secret_accessor(  # type:ignore
            ".".join(config_path)
        )
    return None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MailMessage Entity class."""
from typing import Any, List, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name, too-many-instance-attributes


@export
class MailMessage(Entity):
    """
    MailMessage Entity class.

    Attributes
    ----------
    Recipient : str
        Recipient of the Mail message
    Files : List[Entity]
        Files of the Mail message
    Urls : List[str]
        Urls of the Mail message
    Threats : List[str]
        Threats of the Mail message
    Sender : str
        Sender of the Mail message
    P1Sender : str
        P1Sender of the Mail message
    P1SenderDisplayName : str
        P1SenderDisplayName of the Mail message
    P1SenderDomain : str
        P1SenderDomain of the Mail message
    SenderIP : str
        SenderIP of the Mail message
    P2Sender : str
        P2Sender of the Mail message
    P2SenderDisplayName : str
        P2SenderDisplayName of the Mail message
    P2SenderDomain : str
        P2SenderDomain of the Mail message
    ReceivedDate : datetime
        ReceivedDate of the Mail message
    NetworkMessageId : str
        NetworkMessageId of the Mail message
    InternetMessageId : str
        InternetMessageId of the Mail message
    Subject : str
        Subject of the Mail message
    BodyFingerprintBin1 : str
        BodyFingerprintBin1 of the Mail message
    BodyFingerprintBin2 : str
        BodyFingerprintBin2 of the Mail message
    BodyFingerprintBin3 : str
        BodyFingerprintBin3 of the Mail message
    BodyFingerprintBin4 : str
        BodyFingerprintBin4 of the Mail message
    BodyFingerprintBin5 : str
        BodyFingerprintBin5 of the Mail message
    AntispamDirection : str
        AntispamDirection of the Mail message
    DeliveryAction : str
        DeliveryAction of the Mail message
    DeliveryLocation : str
        DeliveryLocation of the Mail message
    Language : str
        Language of the Mail message
    ThreatDetectionMethods : str
        ThreatDetectionMethods of the Mail message

    """

    ID_PROPERTIES = ["NetworkMessageId", "Recipient"]

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.Recipient: Optional[str] = None
        self.Files: List[Entity] = []
        self.Urls: List[str] = []
        self.Threats: List[str] = []
        self.Sender: Optional[str] = None
        self.P1Sender: Optional[str] = None
        self.P1SenderDisplayName: Optional[str] = None
        self.P1SenderDomain: Optional[str] = None
        self.SenderIP: Optional[str] = None
        self.P2Sender: Optional[str] = None
        self.P2SenderDisplayName: Optional[str] = None
        self.P2SenderDomain: Optional[str] = None
        self.ReceivedDate: Any = None
        self.NetworkMessageId: Optional[str] = None
        self.InternetMessageId: Optional[str] = None
        self.Subject: Optional[str] = None
        self.BodyFingerprintBin1: Optional[str] = None
        self.BodyFingerprintBin2: Optional[str] = None
        self.BodyFingerprintBin3: Optional[str] = None
        self.BodyFingerprintBin4: Optional[str] = None
        self.BodyFingerprintBin5: Optional[str] = None
        self.AntispamDirection: Optional[str] = None
        self.DeliveryAction: Optional[str] = None
        self.DeliveryLocation: Optional[str] = None
        self.Language: Optional[str] = None
        self.ThreatDetectionMethods: Optional[str] = None

        super().__init__(src_entity=src_entity, **kwargs)
        if src_event:
            self._create_from_event(src_event)

    def _create_from_event(self, src_event):
        self.Recipient = src_event.get("Recipient")
        self.Files = src_event.get("FileEntityIds")
        self.Urls = src_event.get("Urls")
        self.Threats = src_event.get("Threats")
        self.SenderIP = src_event.get("SenderIP")
        self.P1Sender = src_event.get("P1Sender")
        self.P1SenderDisplayName = src_event.get("P1SenderDisplayName")
        self.P1SenderDomain = src_event.get("P1SenderDomain")
        self.P2Sender = src_event.get("P2Sender")
        self.P2SenderDisplayName = src_event.get("P2SenderDisplayName")
        self.P2SenderDomain = src_event.get("P2SenderDomain")
        self.ReceivedDate = src_event.get("ReceiveDate")
        self.NetworkMessageId = src_event.get("NetworkMessageId")
        self.InternetMessageId = src_event.get("InternetMessageId")
        self.Subject = src_event.get("Subject")
        self.AntispamDirection = src_event.get("AntispamDirection")
        self.DeliveryAction = src_event.get("DeliveryAction")
        self.Language = src_event.get("Language")

    @property
    def description_str(self):
        """Return Entity Description."""
        return self.NetworkMessageId or self.__class__.__name__

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return (
            self.Subject
            or f"MailMessage to: {self.Recipient}"
            or self.__class__.__name__
        )

    _entity_schema = {
        "Recipient": None,
        "Files": (list, "File"),
        "Urls": None,
        "Threats": None,
        "Sender": None,
        "P1Sender": None,
        "P1SenderDisplayName": None,
        "P1SenderDomain": None,
        "SenderIP": None,
        "P2Sender": None,
        "P2SenderDisplayName": None,
        "P2SenderDomain": None,
        "ReceivedDate": None,
        "NetworkMessageId": None,
        "InternetMessageId": None,
        "Subject": None,
        "BodyFingerprintBin1": None,
        "BodyFingerprintBin2": None,
        "BodyFingerprintBin3": None,
        "BodyFingerprintBin4": None,
        "BodyFingerprintBin5": None,
        "AntispamDirection": None,
        "DeliveryAction": None,
        "DeliveryLocation": None,
        "Language": None,
        "ThreatDetectionMethods": None,
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Module for IoCExtract class.

Uses a set of builtin regular expressions to look for Indicator of
Compromise (IoC) patterns. Input can be a single string or a pandas
dataframe with one or more columns specified as input.

The following types are built-in:

-  IPv4 and IPv6
-  URL
-  DNS domain
-  Hashes (MD5, SHA1, SHA256)
-  Windows file paths
-  Linux file paths (this is kind of noisy because a legal linux file
   path can have almost any character) You can modify or add to the
   regular expressions used at runtime.

"""

import re
from collections import defaultdict, namedtuple
from enum import Enum
from typing import Any, Dict, List, Set, Tuple, Union
from urllib.parse import unquote

import pandas as pd

from .._version import VERSION
from ..common.utility import check_kwargs, export
from .domain_utils import DomainValidator

__version__ = VERSION
__author__ = "Ian Hellen"


def _compile_regex(regex):
    return re.compile(regex, re.I | re.X | re.M)


IoCPattern = namedtuple("IoCPattern", ["ioc_type", "comp_regex", "priority", "group"])

_RESULT_COLS = ["IoCType", "Observable", "SourceIndex", "Input"]


@export
class IoCType(Enum):
    """Enumeration of IoC Types."""

    # pylint: disable=invalid-name
    unknown = "unknown"
    ipv4 = "ipv4"
    ipv6 = "ipv6"
    dns = "dns"
    url = "url"
    md5_hash = "md5_hash"
    sha1_hash = "sha1_hash"
    sha256_hash = "sha256_hash"
    file_hash = "file_hash"
    email = "email"
    windows_path = "windows_path"
    linux_path = "linux_path"
    hostname = "hostname"

    # pylint: enable=invalid-name

    @classmethod
    def parse(cls, value: str) -> "IoCType":
        """
        Return parsed IoCType of string.

        Parameters
        ----------
        value : str
            Enumeration name

        Returns
        -------
        IoCType
            IoCType matching name or unknown if no match

        """
        try:
            ioc_type = IoCType(value.lower())
        except ValueError:
            ioc_type = IoCType.unknown
        return ioc_type


# pylint: enable=invalid-name


@export
class IoCExtract:
    """
    IoC Extractor - looks for common IoC patterns in input strings.

    The extract() method takes either a string or a pandas DataFrame
    as input. When using the string option as an input extract will
    return a dictionary of results. When using a DataFrame the results
    will be returned as a new DataFrame with the following columns:
    IoCType: the mnemonic used to distinguish different IoC Types
    Observable: the actual value of the observable
    SourceIndex: the index of the row in the input DataFrame from
    which the source for the IoC observable was extracted.

    The class has a number of built-in IoC regex definitions.
    These can be retrieved using the ioc_types attribute.

    Addition IoC definitions can be added using the add_ioc_type
    method.

    Note: due to some ambiguity in the regular expression patterns
    for different types and observable may be returned assigned to
    multiple observable types. E.g. 192.168.0.1 is a also a legal file
    name in both Linux and Windows. Linux file names have a particularly
    large scope in terms of legal characters so it will be quite common
    to see other IoC observables (or parts of them) returned as a
    possible linux path.
    """

    IPV4_REGEX = r"(?P<ipaddress>(?:[0-9]{1,3}\.){3}[0-9]{1,3})"
    IPV6_REGEX = r"(?<![:.\w])(?:[A-F0-9]{0,4}:){2,7}[A-F0-9]{0,4}(?![:.\w])"
    DNS_REGEX = r"((?=[a-z0-9-]{1,63}\.)[a-z0-9]+(-[a-z0-9]+)*\.){1,126}[a-z]{2,63}"

    URL_REGEX = r"""
            (?P<protocol>(https?|ftp|telnet|ldap|file)://)
            (?P<userinfo>([a-z0-9-._~!$&\'()*+,;=:]|%[0-9A-F]{2})*@)?
            (?P<host>([a-z0-9-._~!$&\'()*+,;=]|%[0-9A-F]{2})*)
            (:(?P<port>\d*))?
            (/(?P<path>([^?\#"<>\s]|%[0-9A-F]{2})*/?))?
            (\?(?P<query>([a-z0-9-._~!$&'()*+,;=:/?@]|%[0-9A-F]{2})*))?
            (\#(?P<fragment>([a-z0-9-._~!$&'()*+,;=:/?@]|%[0-9A-F]{2})*))?"""

    WINPATH_REGEX = r"""
            (?P<root>[a-z]:|\\\\[a-z0-9_.$-]+||[.]+)
            (?P<folder>\\(?:[^\/:*?"\'<>|\r\n]+\\)*)
            (?P<file>[^\\/*?""<>|\r\n ]+)"""
    # Linux simplified - this ignores some legal linux paths avoid matching too much
    # This also matches URLs but these should be thrown out by priority
    # weighting since URL has a higher priority
    LXPATH_REGEX = r"""(?P<root>/+||[.]+)
            (?P<folder>/(?:[^\\/:*?<>|\r\n]+/)*)
            (?P<file>[^/\0<>|\r\n ]+)"""

    MD5_REGEX = r"(?:^|[^A-Fa-f0-9])(?P<hash>[A-Fa-f0-9]{32})(?:$|[^A-Fa-f0-9])"
    SHA1_REGEX = r"(?:^|[^A-Fa-f0-9])(?P<hash>[A-Fa-f0-9]{40})(?:$|[^A-Fa-f0-9])"
    SHA256_REGEX = r"(?:^|[^A-Fa-f0-9])(?P<hash>[A-Fa-f0-9]{64})(?:$|[^A-Fa-f0-9])"

    _content_regex: Dict[str, IoCPattern] = {}

    def __init__(self):
        """Initialize new instance of IoCExtract."""
        # IP Addresses
        self.add_ioc_type(IoCType.ipv4.name, self.IPV4_REGEX, 0, "ipaddress")
        self.add_ioc_type(IoCType.ipv6.name, self.IPV6_REGEX, 0)

        # Dns Domains
        # This also matches IP addresses but IPs have higher
        # priority both matching on the same substring will defer
        # to the IP regex
        self.add_ioc_type(IoCType.dns.name, self.DNS_REGEX, 1)

        # Http requests
        self.add_ioc_type(IoCType.url.name, self.URL_REGEX, 0)

        # File paths
        # Windows
        self.add_ioc_type(IoCType.windows_path.name, self.WINPATH_REGEX, 2)

        self.add_ioc_type(IoCType.linux_path.name, self.LXPATH_REGEX, 2)

        # MD5, SHA1, SHA256 hashes
        self.add_ioc_type(IoCType.md5_hash.name, self.MD5_REGEX, 1, "hash")
        self.add_ioc_type(IoCType.sha1_hash.name, self.SHA1_REGEX, 1, "hash")
        self.add_ioc_type(IoCType.sha256_hash.name, self.SHA256_REGEX, 1, "hash")

        self._dom_validator = DomainValidator()
        self._ignore_tld = False

    # Public members
    def add_ioc_type(
        self, ioc_type: str, ioc_regex: str, priority: int = 0, group: str = None
    ):
        """
        Add an IoC type and regular expression to use to the built-in set.

        Parameters
        ----------
        ioc_type : str
            A unique name for the IoC type
        ioc_regex : str
            A regular expression used to search for the type
        priority : int, optional
            Priority of the regex match vs. other ioc_patterns. 0 is
            the highest priority (the default is 0).
        group : str, optional
            The regex group to match (the default is None,
            which will match on the whole expression)

        Notes
        -----
        Pattern priorities.
            If two IocType patterns match on the same substring, the matched
            substring is assigned to the pattern/IocType with the highest
            priority. E.g. `foo.bar.com` will match types: `dns`, `windows_path`
            and `linux_path` but since `dns` has a higher priority, the expression
            is assigned to the `dns` matches.

        """
        if ioc_type is None or ioc_type.strip() is None:
            raise Exception("No value supplied for ioc_type parameter")
        if ioc_regex is None or ioc_regex.strip() is None:
            raise Exception("No value supplied for ioc_regex parameter")

        self._content_regex[ioc_type] = IoCPattern(
            ioc_type=ioc_type,
            comp_regex=_compile_regex(regex=ioc_regex),
            priority=priority,
            group=group,
        )

    @property
    def ioc_types(self) -> dict:
        """
        Return the current set of IoC types and regular expressions.

        Returns
        -------
        dict
            dict of IoC Type names and regular expressions

        """
        return self._content_regex

    # pylint: disable=too-many-locals
    def extract(
        self,
        src: str = None,
        data: pd.DataFrame = None,
        columns: List[str] = None,
        **kwargs,
    ) -> Union[Dict[str, Set[str]], pd.DataFrame]:
        """
        Extract IoCs from either a string or pandas DataFrame.

        Parameters
        ----------
        src : str, optional
            source string in which to look for IoC patterns
            (the default is None)
        data : pd.DataFrame, optional
            input DataFrame from which to read source strings
            (the default is None)
        columns : list, optional
            The list of columns to use as source strings,
            if the `data` parameter is used. (the default is None)

        Other Parameters
        ----------------
        ioc_types : list, optional
            Restrict matching to just specified types.
            (default is all types)
        include_paths : bool, optional
            Whether to include path matches (which can be noisy)
            (the default is false - excludes 'windows_path'
            and 'linux_path'). If `ioc_types` is specified
            this parameter is ignored.
        ignore_tlds : bool, optional
            If True, ignore the official Top Level Domains
            list when determining whether a domain name is
            a legal domain.

        Returns
        -------
        Any
            dict of found observables (if input is a string) or
            DataFrame of observables

        Notes
        -----
        Extract takes either a string or a pandas DataFrame as input.
        When using the string option as an input extract will
        return a dictionary of results.
        When using a DataFrame the results will be returned as a new
        DataFrame with the following columns:
        - IoCType: the mnemonic used to distinguish different IoC Types
        - Observable: the actual value of the observable
        - SourceIndex: the index of the row in the input DataFrame from
        which the source for the IoC observable was extracted.

        IoCType Pattern selection
        The default list is:  ['ipv4', 'ipv6', 'dns', 'url',
        'md5_hash', 'sha1_hash', 'sha256_hash'] plus any
        user-defined types.
        'windows_path', 'linux_path' are excluded unless `include_paths`
        is True or explicitly included in `ioc_paths`.

        """
        check_kwargs(kwargs, ["ioc_types", "include_paths", "ignore_tlds"])
        ioc_types = kwargs.get("ioc_types", None)
        include_paths = kwargs.get("include_paths", False)
        ignore_tld_current = self._ignore_tld
        self._ignore_tld = kwargs.get("ignore_tlds", False)

        if src and src.strip():
            return self._scan_for_iocs(src=src, ioc_types=ioc_types)

        if data is None:
            raise Exception("No source data was supplied to extract")

        if columns is None:
            raise Exception("No values were supplied for the columns parameter")

        ioc_types_to_use = self._get_ioc_types_to_use(ioc_types, include_paths)

        col_set = set(columns)
        if col_set > set(data.columns):
            missing_cols = [elem for elem in col_set if elem not in data.columns]
            raise Exception(
                f"Source column(s) {', '.join(missing_cols)} not found",
                " in supplied DataFrame",
            )

        result_rows: List[pd.Series] = []
        for idx, datarow in data.iterrows():
            result_rows.extend(
                self._search_in_row(datarow, idx, columns, ioc_types_to_use)
            )
        self._ignore_tld = ignore_tld_current
        return pd.DataFrame(data=result_rows, columns=_RESULT_COLS)

    # pylint: disable=too-many-arguments
    def _search_in_row(
        self,
        datarow: pd.Series,
        idx: Any,
        columns: List[str],
        ioc_types_to_use: List[str],
    ) -> List[pd.Series]:
        """Return results for a single input row."""
        result_rows = []
        for col in columns:
            ioc_results = self._scan_for_iocs(datarow[col], ioc_types_to_use)
            for result_type, result_set in ioc_results.items():
                if result_set:
                    for observable in result_set:
                        result_row = pd.Series(
                            data=[result_type, observable, idx, datarow[col]],
                            index=_RESULT_COLS,
                        )
                        result_rows.append(result_row)
        return result_rows

    def extract_df(
        self, data: pd.DataFrame, columns: Union[str, List[str]], **kwargs
    ) -> pd.DataFrame:
        """
        Extract IoCs from either a pandas DataFrame.

        Parameters
        ----------
        data : pd.DataFrame
            input DataFrame from which to read source strings
        columns : Union[str, list]
            A single column name as a string or a
            a list of columns to use as source strings,

        Other Parameters
        ----------------
        ioc_types : list, optional
            Restrict matching to just specified types.
            (default is all types)
        include_paths : bool, optional
            Whether to include path matches (which can be noisy)
            (the default is false - excludes 'windows_path'
            and 'linux_path'). If `ioc_types` is specified
            this parameter is ignored.
        ignore_tlds : bool, optional
            If True, ignore the official Top Level Domains
            list when determining whether a domain name is
            a legal domain.

        Returns
        -------
        pd.DataFrame
            DataFrame of observables

        Notes
        -----
        Extract takes a pandas DataFrame as input.
        The results will be returned as a new
        DataFrame with the following columns:
        - IoCType: the mnemonic used to distinguish different IoC Types
        - Observable: the actual value of the observable
        - SourceIndex: the index of the row in the input DataFrame from
        which the source for the IoC observable was extracted.

        IoCType Pattern selection
        The default list is:  ['ipv4', 'ipv6', 'dns', 'url',
        'md5_hash', 'sha1_hash', 'sha256_hash'] plus any
        user-defined types.
        'windows_path', 'linux_path' are excluded unless `include_paths`
        is True or explicitly included in `ioc_paths`.

        """
        check_kwargs(kwargs, ["ioc_types", "include_paths", "ignore_tlds"])
        ioc_types = kwargs.get("ioc_types", None)
        include_paths = kwargs.get("include_paths", False)
        ignore_tld_current = self._ignore_tld
        self._ignore_tld = kwargs.get("ignore_tlds", False)

        ioc_types_to_use = self._get_ioc_types_to_use(ioc_types, include_paths)
        if isinstance(columns, str):
            columns = [columns]
        col_set = set(columns)
        if col_set > set(data.columns):
            missing_cols = [elem for elem in col_set if elem not in data.columns]
            raise Exception(
                f"Source column(s) {', '.join(missing_cols)} not found",
                " in supplied DataFrame",
            )

        result_rows = []
        for idx, datarow in data.iterrows():
            result_rows.extend(
                self._search_in_row(datarow, idx, columns, ioc_types_to_use)
            )
        self._ignore_tld = ignore_tld_current
        return pd.DataFrame(data=result_rows, columns=_RESULT_COLS)

    def _get_ioc_types_to_use(
        self, ioc_types: List[str], include_paths: bool
    ) -> List[str]:
        # Use only requested IoC Type patterns
        if ioc_types:
            ioc_types_to_use = list(set(ioc_types))
        else:
            ioc_types_to_use = list(set(self._content_regex.keys()))
            # don't include linux paths unless explicitly included
            ioc_types_to_use.remove(IoCType.linux_path.name)
            if not include_paths:
                # windows path matching is less noisy
                ioc_types_to_use.remove(IoCType.windows_path.name)
        return ioc_types_to_use

    def validate(
        self, input_str: str, ioc_type: str, ignore_tlds: bool = False
    ) -> bool:
        """
        Check that `input_str` matches the regex for the specificed `ioc_type`.

        Parameters
        ----------
        input_str : str
            the string to test
        ioc_type : str
            the regex pattern to use
        ignore_tlds : bool, optional
            If True, ignore the official Top Level Domains
            list when determining whether a domain name is
            a legal domain.

        Returns
        -------
        bool
            True if match.

        """
        ignore_tld_current = self._ignore_tld
        self._ignore_tld = ignore_tlds
        if ioc_type == IoCType.file_hash.name:
            val_type = self.file_hash_type(input_str).name
        elif ioc_type == IoCType.hostname.name:
            val_type = "dns"
        else:
            val_type = ioc_type
        if val_type not in self._content_regex:
            raise KeyError(
                f"Unknown type {ioc_type}.",
                f"Valid types are: {list(self._content_regex.keys())}",
            )
        rgx = self._content_regex[val_type]
        pattern_match = rgx.comp_regex.fullmatch(input_str)
        validated = self._validate_tld(input_str) if val_type == "dns" else True
        self._ignore_tld = ignore_tld_current
        return pattern_match and validated

    @staticmethod
    def file_hash_type(file_hash: str) -> IoCType:
        """
        Return specific IoCType based on hash length.

        Parameters
        ----------
        file_hash : str
            File hash string

        Returns
        -------
        IoCType
            Specific hash type or unknown.

        """
        hashsize_map = {
            32: IoCType.md5_hash,
            40: IoCType.sha1_hash,
            64: IoCType.sha256_hash,
        }
        hashsize = len(file_hash.strip())
        return hashsize_map.get(hashsize, IoCType.unknown)

    def get_ioc_type(self, observable: str) -> str:
        """
        Return first matching type.

        Parameters
        ----------
        observable : str
            The IoC Observable to check

        Returns
        -------
        str
            The IoC type enumeration (unknown, if no match)

        """
        results = self._scan_for_iocs(src=observable)

        if not results:
            results = self._scan_for_iocs(
                src=observable, ioc_types=[IoCType.linux_path.name]
            )
        if not results:
            return IoCType.unknown.name

        # we need to select the type that is an exact match for the whole
        # observable string (_scan_for_iocs will return matching substrings)
        for ioc_type, match_set in results.items():
            if observable in match_set:
                return ioc_type

        return IoCType.unknown.name

    # Private methods
    def _validate_tld(self, domain: str) -> bool:
        """If validate TLDS check with TLD list."""
        if self._ignore_tld:
            return True
        return self._dom_validator.validate_tld(domain)

    def _scan_for_iocs(
        self, src: str, ioc_types: List[str] = None
    ) -> Dict[str, Set[str]]:
        """Return IoCs found in the string."""
        ioc_results: Dict[str, Set] = defaultdict(set)
        iocs_found: Dict[str, Tuple[str, int]] = {}

        # pylint: disable=too-many-nested-blocks
        for (ioc_type, rgx_def) in self._content_regex.items():
            if ioc_types and ioc_type not in ioc_types:
                continue

            match_pos = 0
            for rgx_match in rgx_def.comp_regex.finditer(src, match_pos):
                if rgx_match is None:
                    break
                # If the rgx_def names a group to match on, use that
                match_str = (
                    rgx_match.groupdict()[rgx_def.group]
                    if rgx_def.group
                    else rgx_match.group()
                )

                if ioc_type == "dns" and not self._validate_tld(match_str):
                    continue

                self._add_highest_pri_match(iocs_found, match_str, rgx_def)
                if ioc_type == "url":
                    self._check_decode_url(match_str, rgx_def, match_pos, iocs_found)
                match_pos = rgx_match.end()

        for ioc, ioc_result in iocs_found.items():
            ioc_results[ioc_result[0]].add(ioc)

        return ioc_results

    def _check_decode_url(self, match_str, rgx_def, match_pos, iocs_found):
        """Get any other IoCs from decoded URL."""
        decoded_url = unquote(match_str)
        for url_match in rgx_def.comp_regex.finditer(decoded_url, match_pos):
            if url_match is not None:
                self._add_highest_pri_match(iocs_found, url_match.group(), rgx_def)
                self._add_highest_pri_match(
                    iocs_found,
                    url_match.groupdict()["host"],
                    self._content_regex["dns"],
                )

    @staticmethod
    def _add_highest_pri_match(
        iocs_found: dict, current_match: str, current_def: IoCPattern
    ):
        # if we already found a match for this item and the previous
        # ioc type is more specific then don't add this to the results
        if (
            current_match in iocs_found
            and current_def.priority >= iocs_found[current_match][1]
        ):
            return

        iocs_found[current_match] = (current_def.ioc_type, current_def.priority)


# pylint: disable=too-few-public-methods
@pd.api.extensions.register_dataframe_accessor("mp_ioc")
class IoCExtractAccessor:
    """Pandas api extension for IoC Extractor."""

    def __init__(self, pandas_obj):
        """Instantiate pandas extension class."""
        self._df = pandas_obj
        self._ioc = IoCExtract()

    def extract(self, columns, **kwargs):
        """
        Extract IoCs from either a pandas DataFrame.

        Parameters
        ----------
        columns : list
            The list of columns to use as source strings,

        Other Parameters
        ----------------
        ioc_types : list, optional
            Restrict matching to just specified types.
            (default is all types)
        include_paths : bool, optional
            Whether to include path matches (which can be noisy)
            (the default is false - excludes 'windows_path'
            and 'linux_path'). If `ioc_types` is specified
            this parameter is ignored.

        Returns
        -------
        pd.DataFrame
            DataFrame of observables

        Notes
        -----
        Extract takes a pandas DataFrame as input.
        The results will be returned as a new
        DataFrame with the following columns:
        - IoCType: the mnemonic used to distinguish different IoC Types
        - Observable: the actual value of the observable
        - SourceIndex: the index of the row in the input DataFrame from
        which the source for the IoC observable was extracted.

        IoCType Pattern selection
        The default list is:  ['ipv4', 'ipv6', 'dns', 'url',
        'md5_hash', 'sha1_hash', 'sha256_hash'] plus any
        user-defined types.
        'windows_path', 'linux_path' are excluded unless `include_paths`
        is True or explicitly included in `ioc_paths`.

        """
        return self._ioc.extract_df(data=self._df, columns=columns, **kwargs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Uses the Azure Python SDK to interact with Azure Blob Storage."""
import datetime
from typing import Any, List, Optional

import pandas as pd
from azure.common.exceptions import CloudError
from azure.core.exceptions import (
    ResourceExistsError,
    ResourceNotFoundError,
    ServiceRequestError,
)
from azure.storage.blob import BlobServiceClient, generate_blob_sas

from ..common.azure_auth import az_connect
from ..common.azure_auth_core import AzCredentials


class AzureBlobStorage:
    """Class for interacting with Azure Blob Storage."""

    def __init__(self, abs_name: str, connect: bool = False):
        """Initialize connector for Azure Python SDK."""
        self.connected = False
        self.abs_site = f"{abs_name}.blob.core.windows.net"
        self.credentials: Optional[AzCredentials] = None
        self.abs_client: Optional[BlobServiceClient] = None
        if connect is True:
            self.connect()

    def connect(
        self,
        auth_methods: List = None,
        silent: bool = False,
    ):
        """Authenticate with the SDK."""
        self.credentials = az_connect(auth_methods=auth_methods, silent=silent)
        if not self.credentials:
            raise CloudError("Could not obtain credentials.")
        self.abs_client = BlobServiceClient(self.abs_site, self.credentials.modern)
        if not self.abs_client:
            raise CloudError("Could not create a Blob Storage client.")
        self.connected = True

    def containers(self) -> pd.DataFrame:
        """Return containers in the Azure Blob Storage Account."""
        try:
            container_list = self.abs_client.list_containers()  # type:ignore
        except ServiceRequestError as err:
            raise CloudError(
                "Unable to connect check the Azure Blob Store account name"
            ) from err
        if container_list:
            containers_df = _parse_returned_items(
                container_list, remove_list=["lease", "encryption_scope"]
            )
        else:
            containers_df = None
        return containers_df

    def create_container(self, container_name: str, **kwargs) -> pd.DataFrame:
        """
        Create a new container within the Azure Blob Storage account.

        Parameters
        ----------
        container_name : str
            The name for the new container.
        Additional container parameters can be passed as kwargs

        Returns
        -------
        pd.DataFrame
            Details of the created container.

        """
        try:
            new_container = self.abs_client.create_container(  # type: ignore
                container_name, **kwargs
            )  # type:ignore
        except ResourceExistsError as err:
            raise CloudError(f"Container {container_name} already exists.") from err
        properties = new_container.get_container_properties()
        container_df = _parse_returned_items(
            [properties], ["encryption_scope", "lease"]
        )
        return container_df

    def blobs(self, container_name: str) -> Optional[pd.DataFrame]:
        """
        Get a list of blobs in a container.

        Parameters
        ----------
        container_name : str
            The name of the container to get blobs from.

        Returns
        -------
        pd.DataFrame
            Details of the blobs.

        """
        container_client = self.abs_client.get_container_client(container_name)  # type: ignore
        blobs = list(container_client.list_blobs())
        return _parse_returned_items(blobs) if blobs else None

    def upload_to_blob(
        self, blob: Any, container_name: str, blob_name: str, overwrite: bool = True
    ):
        """
        Upload a blob of data.

        Parameters
        ----------
        blob : Any
            The data to upload.
        container_name : str
            The name of the container to upload the blob to.
        blob_name : str
            The name to give the blob.
        overwrite : bool, optional
            Whether or not you want to overwrite the blob if it exists, by default True.

        """
        try:
            blob_client = self.abs_client.get_blob_client(  # type:ignore
                container=container_name, blob=blob_name
            )
            upload = blob_client.upload_blob(blob, overwrite=overwrite)
        except ResourceNotFoundError as err:
            raise CloudError(
                "Unknown container, check container name or create it first."
            ) from err
        if not upload["error_code"]:
            print("Upload complete")
        else:
            raise CloudError(
                f"There was a problem uploading the blob: {upload['error_code']}"
            )
        return True

    def get_blob(self, container_name: str, blob_name: str) -> bytes:
        """
        Get a blob from the Azure Blob Storage account.

        Parameters
        ----------
        container_name : str
            The name of the container that holds the blob.
        blob_name : str
            The name of the blob to download.

        Returns
        -------
        bytes
            The content of the blob in bytes.

        """
        blob_client = self.abs_client.get_blob_client(  # type: ignore
            container=container_name, blob=blob_name
        )
        if blob_client.exists():
            data_stream = blob_client.download_blob()
            data = data_stream.content_as_bytes()
        else:
            raise CloudError(f"The blob {blob_name} does not exist in {container_name}")
        return data

    def delete_blob(self, container_name: str, blob_name: str) -> bool:
        """
        Delete a blob from the Azure Blob Storage account.

        Parameters
        ----------
        container_name : str
            The container name that has the blob.
        blob_name : str
            The name of the blob to delete.
        Note deleting a blob also deletes associated snapshots.

        Returns
        -------
        bool
            True if blob successfully deleted

        """
        blob_client = self.abs_client.get_blob_client(  # type: ignore
            container=container_name, blob=blob_name
        )
        if blob_client.exists():
            blob_client.delete_blob(delete_snapshots="include")
        else:
            raise CloudError(f"The blob {blob_name} does not exist in {container_name}")

        return True

    def get_sas_token(
        self,
        container_name: str,
        blob_name: str,
        end: datetime.datetime = None,
        permission: str = "r",
    ) -> str:
        """
        Generate a shared access string (SAS) token for a blob.

        Parameters
        ----------
        container_name : str
            The name of the Azure Blob Storage container that holds the blob.
        blob_name : str
            The name of the blob to generate the SAS token for.
        end : datetime.datetime, optional
            The datetime the SAS token should expire, by default this is 7 days from now.
        permission : str, optional
            The permissions to give the SAS token, by default 'r' for read.

        Returns
        -------
        str
            A URI of the blob with SAS token.

        """
        start = datetime.datetime.now()
        if not end:
            end = start + datetime.timedelta(days=7)
        key = self.abs_client.get_user_delegation_key(start, end)  # type: ignore
        abs_name = self.abs_client.account_name  # type: ignore
        sast = generate_blob_sas(
            abs_name,
            container_name,
            blob_name,
            user_delegation_key=key,
            permission=permission,
            expiry=end,
            start=start,
        )
        full_path = f"https://{abs_name}.blob.core.windows.net/{container_name}/{blob_name}?{sast}"
        return full_path


def _parse_returned_items(items, remove_list: list = None) -> pd.DataFrame:
    """Parse a list of containers into a DataFrame."""
    out_items = []
    for item in items:
        item = dict(item)
        if remove_list:
            for remove_item in remove_list:
                item.pop(remove_item)
        out_items.append(item)
    return pd.json_normalize(out_items)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Tor Exit Nodes Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from datetime import datetime
from threading import Lock
from typing import Tuple, Iterable, Dict, Any

import requests

from .ti_provider_base import TIProvider, LookupResult, TISeverity, TILookupStatus
from ...common.utility import export
from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class Tor(TIProvider):
    """Tor Exit Nodes Lookup."""

    _BASE_URL = "https://check.torproject.org/exit-addresses"

    _IOC_QUERIES: dict = {"ipv4": None}
    _nodelist: Dict[str, Dict[str, str]] = {}
    _last_cached = datetime.min
    _cache_lock = Lock()

    def __init__(self, **kwargs):
        """Instantiate Tor class."""
        super().__init__(**kwargs)
        self._check_and_get_nodelist()

    @classmethod
    def _check_and_get_nodelist(cls):
        """Pull down Tor exit node list and save to internal attribute."""
        if cls._cache_lock.locked():
            return
        now = datetime.utcnow()
        if not cls._nodelist or (now - cls._last_cached).days > 1:
            try:
                resp = requests.get(cls._BASE_URL)
                tor_raw_list = resp.content.decode()
                with cls._cache_lock:
                    cls._nodelist = dict(cls._tor_splitter(tor_raw_list))
                    cls._last_cached = datetime.utcnow()
            except ConnectionError:
                pass

    @staticmethod
    def _tor_splitter(node_list) -> Iterable[Tuple[str, Dict[str, str]]]:
        node_dict: Dict[str, str] = {}
        for line in node_list.split("\n"):
            if not line:
                continue
            fields = line.split(" ", 2)
            if fields[0] == "ExitNode":
                # new record so reset dict
                node_dict = {}
            node_dict[fields[0]] = fields[1] if len(fields) > 1 else None
            if fields[0] == "ExitAddress":
                # yield tuple
                yield fields[1], node_dict

    def lookup_ioc(
        self, ioc: str, ioc_type: str = None, query_type: str = None, **kwargs
    ) -> LookupResult:
        """
        Lookup a single IoC observable.

        Parameters
        ----------
        ioc : str
            IoC Observable value
        ioc_type : str, optional
            IoC Type, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.

        Returns
        -------
        LookupResult
            The returned results.

        """
        result = self._check_ioc_type(
            ioc=ioc, ioc_type=ioc_type, query_subtype=query_type
        )

        result.provider = kwargs.get("provider_name", self.__class__.__name__)
        result.result = bool(self._nodelist)
        result.reference = self._BASE_URL

        if result.status and not bool(self._nodelist):
            result.status = TILookupStatus.query_failed.value

        if result.status:
            return result

        tor_node = self._nodelist.get(ioc)

        if tor_node:
            result.set_severity(TISeverity.warning)
            result.details = {
                "NodeID": tor_node["ExitNode"],
                "LastStatus": tor_node["LastStatus"],
            }
            result.raw_result = tor_node
        else:
            result.details = "Not found."
        return result

    def parse_results(self, response: LookupResult) -> Tuple[bool, TISeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : LookupResult
            The returned data response

        Returns
        -------
        Tuple[bool, TISeverity, Any]
            bool = positive or negative hit
            TISeverity = enumeration of severity
            Object with match details

        """
        return (True, TISeverity.information, None)
